package ctalz.cnstr;

import java.io.*;
import static ctalz.cnstr.CVCParser.*;

%%

%class CVCLexer
%implements CVCParser.yyInput
%unicode
%type Token

%{

private Token token;

public boolean advance() throws IOException
{
	token = yylex();
	if (token == null)
		return false;
	return true;
}
public int token()
{
	return token.type;
}
public Token value()
{
	return token;
}

%}

SP = [ \t\n\r]
ID = [a-zA-Z_][a-zA-Z0-9_$:]*

%%

<YYINITIAL> {
  {ID}                          { return new Token(ID, yytext()); }
  "-"? [0-9]+                   {
    String t = yytext();
    try {
      return new Token(INT_LITERAL, Integer.parseInt(t));
    } catch (NumberFormatException e) {
      throw new RuntimeException();
    }
  }
  {SP}+                         { }
  [\x00-\xFF]                   { return new Token((int)yytext().charAt(0)); }
  .                             { throw new RuntimeException(); }
}
