package ctalz.cnstr;

import java.util.*;
import ctalz.util.*;

public class CombProp extends Prop
{
	private Op op;
	private Prop left, right;
	
	public static enum Op
	{
		AND("AND", "and", 2, 1), OR("OR", "or", 2, 2), NOT("NOT", "not", 1, 3), 
		IMP("=>", "imp", 2, 10), EQUIV("<=>", "equiv", 2, 10);
		
		private String solverName, userName;
		private int arity, id;
		Op(String p, String u, int a, int i) { solverName = p; userName = u; arity = a; id = i; }
	}
	
	public CombProp(Op o, Prop l)
	{
		assert o.arity == 1;
		op = o; left = l;
	}
	public CombProp(Op o, Prop l, Prop r)
	{
		assert o.arity == 2;
		op = o; left = l; right = r;
	}
	
	protected Collection<VarExpr> collectFreeVariables(Collection<VarExpr> set)
	{
		left.collectFreeVariables(set);
		if (right != null)
			right.collectFreeVariables(set);
		return set;
	}
	
	public Prop substitute(Map<VarExpr, Expr> map)
	{
		Prop newLeft = left.substitute(map);
		if (right != null) {
			Prop newRight = right.substitute(map);
			return (newLeft == left && newRight == right ? this : new CombProp(op, newLeft, newRight));
		} else
			return (newLeft == left ? this : new CombProp(op, newLeft));
	}
	
	protected void appendForUser(StringBuilder out, int outerOp)
	{
		if (outerOp != 0 && (op.id != outerOp || op.id == 10))
			out.append('(');
		if (right != null) {
			left.appendForUser(out, op.id);
			out.append(' ').append(op.userName).append(' ');
			right.appendForUser(out, op.id);
		} else {
			out.append(op.userName).append(' ');
			left.appendForUser(out, op.id);
		}
		if (outerOp != 0 && (op.id != outerOp || op.id == 10))
			out.append(')');
	}
	protected void appendForSolver(StringBuilder out)
	{
		if (right != null) {
			out.append('(');
			left.appendForSolver(out); out.append(op.solverName); right.appendForSolver(out);
			out.append(')');
		} else {
			out.append('(');
			out.append(op.solverName); left.appendForSolver(out);
			out.append(')');
		}
	}
	
	public Boolean evaluate()
	{
		Boolean lv = left.evaluate();
		if (lv == null)
			return null;
		if (op == Op.NOT)
			return !lv;
		Boolean rv = right.evaluate();
		if (rv == null)
			return null;
		switch (op) {
		case AND: return (lv & rv);
		case OR: return (lv | rv);
		case IMP: return (!lv | rv);
		case EQUIV: return (lv == rv);
		}
		return null;
	}
	public boolean deepEquals(Prop prop)
	{
		if (!(prop instanceof CombProp))
			return false;
		CombProp p = (CombProp)prop;
		if (op != p.op || (right == null) != (p.right == null))
			return false;
		return (left.deepEquals(p.left) && (right == null || right.deepEquals(p.right)));
	}
}
