package ctalz.cnstr;

import java.util.*;
import ctalz.util.*;

public abstract class ConstraintSolver
{
	protected static long timeSpentInSolver;
	
	public static ConstraintSolver newInstance()
	{
		try {
			return (ConstraintSolver)Thread.currentThread().getContextClassLoader()
										 .loadClass(Config.asString("ConstraintSolverClass", "")).newInstance();
		} catch (Exception e) {
			ExceptionAccumulator.addFatal(Location.NULL, "cannotLaunchSolver", e);
			return null;
		}
	}
	
	public abstract void terminate();
	
	public static long nanoTimeSpentInSolver()
	{
		return timeSpentInSolver;
	}
	
	public abstract void reset();
	public abstract void saveEnvironment();
	public abstract void restoreEnvironment();
	
	public abstract void assume(Prop prop);
	public abstract boolean isValid(Prop prop);
	public boolean isUnsatisfiable(Prop prop)
	{
		return isValid(new CombProp(CombProp.Op.NOT, prop));
	}
	public boolean isEquivalent(Expr e1, Expr e2)
	{
		return (e1 == e2 || e1.deepEquals(e2) || isValid(new RelProp(RelProp.Op.EQ, e1, e2)));
	}
	public boolean isEquivalent(Prop p1, Prop p2)
	{
		return (p1 == p2 || p1.deepEquals(p2) || isValid(new CombProp(CombProp.Op.EQUIV, p1, p2)));
	}
	public abstract Expr simplify(Expr expr);
	public abstract Prop simplify(Prop prop);
}
