package ctalz.comp.act;

import ctalz.comp.ctype.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class ActType
{
	public final Kind kind;
	public final Expr val, len, meta;
	public final LabelType label;
	
	public enum Kind
	{
		INT("int", 4), BP("base pointer", 4), PACKED_BP("packed base pointer", 4), 
		RET_ADDR("ret", 4), VAR("var", 4);
		
		final String name;
		final int size;
		Kind(String n, int s) { name = n; size = s; }
		public String toString() { return name; }
	}
	
	private ActType(Kind k, Expr v, Expr l, Expr m, LabelType b)
	{
		kind = k; val = v; len = l; meta = m; label = b;
	}
	public static ActType newInt(Expr v)
	{
		return new ActType(Kind.INT, v, null, null, null);
	}
	public static ActType newInt(int v)
	{
		return new ActType(Kind.INT, ConstExpr.valueOf(v), null, null, null);
	}
	public static ActType freshInt()
	{
		return newInt(new VarExpr("v" + Utilities.uniqueID()));
	}
	public static ActType newPackedBP(Expr l, Expr m)
	{
		if (l == null && m == null)
			return freshPackedBP();
		ActType t = new ActType(Kind.PACKED_BP, null, l, m, null);
		return t;
	}
	public static ActType freshPackedBP()
	{
		return newPackedBP(new VarExpr("l" + Utilities.uniqueID()), new VarExpr("m" + Utilities.uniqueID()));
	}
	public static ActType newBP(Expr v, Expr l, Expr m)
	{
		return new ActType(Kind.BP, v, l, m, null);
	}
	public static ActType newRetAddr(LabelType type)
	{
		return new ActType(Kind.RET_ADDR, null, null, null, type);
	}
	public static ActType newVar(String name)
	{
		return new ActType(Kind.VAR, new VarExpr(name), null, null, null);
	}
	public static ActType fromCType(CType t)
	{
		t = t.dequalify();
		if (t.isIntegral() || t.isPtr())
			return freshInt();
		assert false;
		return freshInt();
	}
	
	public int sizeof()
	{
		return kind.size;
	}
	
	@Override public boolean equals(Object obj)
	{
		if (!(obj instanceof ActType))
			return false;
		ActType t = (ActType)obj;
		return (kind == t.kind && (val == null ? t.val == null : val.equals(t.val)) && 
				(len == null ? t.len == null : len.equals(t.len)) && 
				(meta == null ? t.meta == null : meta.equals(t.meta)));
	}
	@Override public int hashCode()
	{
		return kind.hashCode() + (val == null ? 0 : val.hashCode()) + (len == null ? 0 : len.hashCode());
	}
	
	public String toString()
	{
		switch (kind) {
		case INT: return "I<" + val.toShortString() + ">";
		case BP: return "B<" + len.toShortString() + "," + meta.toShortString() + ">";
		case PACKED_BP: return "PB<" + len.toShortString() + "," + meta.toShortString() + ">";
		case RET_ADDR: return "R";
		case VAR: return "T";
		default: assert false; return "";
		}
	}
	String toAsmString()
	{
		switch (kind) {
		case INT: return "int(" + val.toShortString() + ")";
		case PACKED_BP: return "BasePtr";
		case RET_ADDR: return "ex[r][r=" + label + "][]:int(r)";
		case VAR: return val.toShortString();
		default: assert false; return "";
		}
	}
}
