package ctalz.comp.act;

import java.io.*;
import java.util.*;
import ctalz.util.*;

public class BBlock
{
	private static int dumpACT = Config.asInt("IncludeIntermediateTreeInAssembly", 0);
	
	final Label label;
	List<ActNode> nodes = new ArrayList<ActNode>();
	
	BBlock(Label l)
	{
		label = l;
		label.bBlock = this;
		nodes.add(label);
	}
	
	void coercionConvert(Program prog, StackLayout stack)
	{
		List<ActNode> newNodes = new ArrayList<ActNode>(nodes.size() * 2);
		RegFile regs = new RegFile();
		stack.resetTypes();
		for (ActNode n : nodes) {
			if (dumpACT > 0) {
				newNodes.add(new Asm("#", n.toString()));
				if (dumpACT > 1) {
					newNodes.add(new Asm("#", regs.toString()));
					if (dumpACT > 2)
						newNodes.add(new Asm("#", stack.toString()));
				}
			}
			n.coercionConvert(prog, newNodes, stack, regs);
		}
		nodes = newNodes;
	}
	void prologueConvert(Program prog, StackLayout stack)
	{
		List<ActNode> newNodes = new ArrayList<ActNode>(nodes.size() * 2);
		RegFile regs = new RegFile();
		stack.resetTypes();
		for (ActNode n : nodes)
			n.prologueConvert(prog, newNodes, stack);
		nodes = newNodes;
	}
	
	void emitAssembly(PrintStream out, BBlock next)
	{
		if (next != null && nodes.get(nodes.size() - 1) instanceof Jump) {
			Jump j = (Jump)nodes.get(nodes.size() - 1);
			if (j.op == Jump.Op.ALWAYS && j.dest == next.label)
				j.isFallThrough = true;
		}
		for (ActNode n : nodes)
			out.println(n instanceof Label ? ((Label)n).toAsmLabelString() : n.toAsmString());
	}
	
	public String toString()
	{
		StringBuilder result = new StringBuilder();
		int count = 0;
		for (ActNode n : nodes) {
			if (count++ > 0)
				result.append(Utilities.EOL);
			result.append(n);
		}
		return result.toString();
	}
}
