package ctalz.comp.act;

import java.io.*;
import java.util.*;
import ctalz.comp.*;
import ctalz.comp.ast.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Function
{
	static PrintStream dumpACT;
	static {
		int p = Config.asInt("PrintIntermediateTree", 0);
		dumpACT = (p == 1 ? System.out : p > 1 ? System.err : null);
	}
	
	private String name;
	private boolean isGlobal;
	private List<Var> args;
	private Var userRet, baseRet;
	private List<ActNode> nodes = new LinkedList<ActNode>();
	private List<BBlock> bBlocks;
	private StackLayout stack;
	
	public Function(String n, boolean g, List<Var> a, Var u, Var b)
	{
		name = n; isGlobal = g; userRet = u; baseRet = b;
		args = new ArrayList<Var>(a);
		Collections.reverse(args);
		nodes.add(new Prologue(true, name + (isGlobal ? "" : Main.uniqueHash()), userRet, baseRet));
	}
	
	public void add(ActNode node)
	{
		nodes.add(node);
	}
	
	public void compile(Program prog)
	{
		if (dumpACT != null) {
			Collections.reverse(args);
			dumpACT.println("Function " + name + "(" + Utilities.enumerate(args, ", ") + "):");
			Collections.reverse(args);
		}
		
		Var retAddr;
		stack = new StackLayout(args);
		retAddr = new Var(ActType.newRetAddr(LabelType.forReturnAddress(stack, userRet, baseRet)), "$retAddr");
		nodes.add(new Prologue(false, name, userRet, baseRet));
		nodes.add(new Return());
		
		buildBBlocks();
		Manipulator.optimize(bBlocks);
		if (dumpACT != null) {
			for (BBlock b : bBlocks)
				dumpACT.println(b);
			dumpACT.println();
		}
		
		stack.find(retAddr); stack.initialElems++;
		for (Var v : stack.calleeSaveVars)
			stack.find(v);
		if (baseRet != null) stack.find(baseRet);
		if (userRet != null) stack.find(userRet);
		
		for (BBlock b : bBlocks)
			b.coercionConvert(prog, stack);
		for (BBlock b : bBlocks)
			b.prologueConvert(prog, stack);
	}
	static Pair<LabelType, LabelType> entryLabelType(List<Var> args, Var userRet, Var baseRet)
	{
		StackLayout s = new StackLayout(args);
		Var ret = new Var(ActType.newRetAddr(LabelType.forReturnAddress(s, userRet, baseRet)), "$retAddr");
		RegFile regs = new RegFile(s);
		s.find(ret);
		return new Pair<LabelType, LabelType>(new LabelType(s, regs), ret.type.label);
	}
	
	void emitAssembly(Program prog, PrintStream out)
	{
		if (isGlobal)
			out.println(".directive \".globl " + Arch.encodeLabelName(name) + "\"");
		
		BBlock prev = null;
		for (BBlock b : bBlocks) {
			if (prev != null)
				prev.emitAssembly(out, b);
			prev = b;
		}
		if (prev != null)
			prev.emitAssembly(out, null);
		out.println();
	}
	
	void buildBBlocks()
	{
		bBlocks = new LinkedList<BBlock>();
		BBlock block = null;
		boolean unreachable = true;
		for (ActNode n : nodes) {
			if (n instanceof Label) {
				if (!unreachable)
					block.nodes.add(new Jump(n));
				block = new BBlock((Label)n);
				bBlocks.add(block);
				unreachable = false;
			} else if (!unreachable) {
				block.nodes.add(n);
				if (n.isUnconditionalJump())
					unreachable = true;
			}
		}
	}
}
