package ctalz.comp.act;

import java.io.*;
import java.util.*;
import ctalz.comp.*;
import ctalz.comp.ast.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Global
{
	final Kind kind;
	final String name, baseName, funName, contents;
	final boolean isBasePtr, isGlobal;
	
	enum Kind
	{
		VAR, GENERIC, STRING;
	}
	
	public Global(String n, boolean b, boolean g)
	{
		kind = Kind.VAR; name = n; baseName = funName = contents = null; isBasePtr = b; isGlobal = g;
	}
	public Global(String f, String n, String s, boolean g)
	{
		kind = Kind.GENERIC; name = n; baseName = s; funName = f; contents = null; isBasePtr = false; isGlobal = g;
	}
	public Global(String c)
	{
		int id = Utilities.uniqueID();
		kind = Kind.STRING; name = ".string" + id; baseName = name + ".b"; funName = null; contents = c;
		isBasePtr = false; isGlobal = false;
	}
	
	void emitDefinition(Program prog, PrintStream out)
	{
		if (Function.dumpACT != null)
			Function.dumpACT.println("Variable " + name);
		if (kind == Kind.VAR)
			emitVariable(prog, out, true);
		else if (kind == Kind.GENERIC)
			emitGeneric(prog, out, true);
		else
			emitString(prog, out, true);
	}
	void emitInitializer(Program prog, PrintStream out)
	{
		if (kind == Kind.VAR)
			emitVariable(prog, out, false);
		else if (kind == Kind.GENERIC)
			emitGeneric(prog, out, false);
		else
			emitString(prog, out, false);
	}
	private void emitVariable(Program prog, PrintStream out, boolean def)
	{
		if (def && isGlobal)
			out.println(".directive \".globl " + name() + "\"");
		if (!isBasePtr)
			out.print(Program.initTemplates[def ? 2 : 3].replace("%name", name()));
		else
			out.print(Program.initTemplates[def ? 4 : 5].replace("%name", name()));
	}
	private void emitGeneric(Program prog, PrintStream out, boolean def)
	{
		if (def && isGlobal) {
			out.println(".directive \".globl " + name() + "\"");
			out.println(".directive \".globl " + baseName() + "\"");
		}
		String cont = ".L" + Utilities.uniqueID();
		out.print(Program.initTemplates[def ? 6 : 7].replace("%user", name()).replace("%base", baseName())
													.replace("%fun", funName()).replace("%cont", cont)
													.replace("%hash", Main.uniqueHash()));
	}
	private void emitString(Program prog, PrintStream out, boolean def)
	{
		byte[] bytes;
		try {
			if (Main.encoding() == null)
				throw new UnsupportedEncodingException();
			bytes = contents.getBytes(Main.encoding());
		} catch (UnsupportedEncodingException e) {
			bytes = contents.getBytes();
		}
		String cont1 = ".L" + Utilities.uniqueID(), cont2 = ".L" + Utilities.uniqueID();
		out.print(Program.initTemplates[def ? 8 : 9]
							  .replace("%user", name()).replace("%base", baseName())
							  .replace("%ascii", name() + ".c").replace("%str", Utilities.escape(bytes) + "\\000")
							  .replace("%len", Integer.toString(bytes.length + 1))
							  .replace("%cont1", cont1).replace("%cont2", cont2)
							  .replace("%malloc", Arch.encodeLabelName(FunDef.mallocName))
							  .replace("%hash", Main.uniqueHash()));
	}
	private String name()
	{
		return Arch.encodeLabelName(name) + (isGlobal ? "" : Main.uniqueHash());
	}
	private String baseName()
	{
		return Arch.encodeLabelName(baseName) + (isGlobal ? "" : Main.uniqueHash());
	}
	private String funName()
	{
		return Arch.encodeLabelName(funName) + (isGlobal ? "" : Main.uniqueHash());
	}
}
