package ctalz.comp.act;

import java.util.*;
import ctalz.comp.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Label extends ActNode
{
	String name;
	boolean isGlobal;
	BBlock bBlock;
	protected StackLayout requiredStack;
	protected RegFile requiredRegs;
	protected LabelType labelType;
	
	public Label(String n)
	{
		name = n;
	}
	public Label(String n, boolean g)
	{
		name = n;
		isGlobal = g;
	}
	
	boolean hasSideEffects()
	{
		return true;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		requiredStack = stack.saveTypes().resetTypes();
		requiredRegs = new RegFile();
		requiredRegs.types[Reg.scratch.index] = ActType.freshInt();
		name = (name == null ? null : Arch.encodeLabelName(name));
		nodes.add(this);
	}
	void prologueConvert(Program prog, List<ActNode> nodes, StackLayout stack)
	{
		nodes.add(this);
	}
	List<Pair<String, String>> prologueConvertForContinuation(StackLayout stack, RegFile regs, Var userRet, Var baseRet)
	{
		List<Pair<String, String>> rename = new LinkedList<Pair<String, String>>();
		for (int i = 1; i <= stack.extraElems; i++) {
			if (stack.types.get(stack.types.size() - i).kind == ActType.Kind.INT)
				stack.types.set(stack.types.size() - i, ActType.freshInt());
		}
		if (userRet != null && userRet.type.kind == ActType.Kind.INT) {
			String oldName = userRet.type.val.toShortString();
			userRet = new Var(ActType.freshInt(), null);
			rename.add(new Pair<String, String>(oldName, userRet.type.val.toShortString()));
		}
		labelType = LabelType.forContinuation(stack, userRet, baseRet);
		return rename;
	}
	
	LabelType labelType()
	{
		if (labelType == null)
			labelType = new LabelType(requiredStack.completeMissingTypes(), requiredRegs);
		return labelType;
	}
	
	public String toString()
	{
		return "(label" + (isGlobal ? "/g" : "") + " " + id + (name == null ? "" : " \"" + name + "\"") + ")";
	}
	String toAsmString()
	{
		return (isGlobal ? name : ".L" + id + (name == null ? "" : "_" + escape(name)) + Main.uniqueHash());
	}
	String toAsmLabelString()
	{
		return toAsmString() + ": " + labelType() + ", []";
	}
	private static String escape(String name)
	{
		return name.replaceAll("[^a-zA-Z0-9_.$]", "_");
	}
}
