package ctalz.comp.act;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class LoadUB extends ActNode
{
	final int size;
	ActNode userDest, baseDest, addr, base;
	
	public LoadUB(int s, ActNode u, ActNode b, ActNode a, ActNode c)
	{
		assert u == null || (u instanceof Reg) || (u instanceof Var);
		assert b == null || (b instanceof Reg) || (b instanceof Var);
		assert s == 4 || b == null;
		size = s; userDest = u; baseDest = b; addr = a; base = c;
	}
	
	void registerVars(VarSet set)
	{
		set.register(userDest); set.register(baseDest); set.register(addr); set.register(base);
	}
	VarSet uses(boolean add, VarSet result)
	{
		if (addr instanceof Var) { if (add) result.add((Var)addr); else result.remove((Var)addr); }
		if (base instanceof Var) { if (add) result.add((Var)base); else result.remove((Var)base); }
		return result;
	}
	VarSet defines(boolean add, VarSet result)
	{
		if (userDest instanceof Var) { if (add) result.add((Var)userDest); else result.remove((Var)userDest); }
		if (baseDest instanceof Var) { if (add) result.add((Var)baseDest); else result.remove((Var)baseDest); }
		return result;
	}
	boolean hasSideEffects()
	{
		return ((userDest instanceof Var && ((Var)userDest).isVolatile) || 
				(baseDest instanceof Var && ((Var)baseDest).isVolatile));
	}
	
	boolean foldConstant(Substitution<ActNode> map, ActNode next)
	{
		if (userDest != null) map.removeKeyValue(userDest);
		if (baseDest != null) map.removeKeyValue(baseDest);
		userDest = Manipulator.swapDest(userDest, next);
		baseDest = Manipulator.swapDest(baseDest, next);
		addr = map.get(addr); base = map.get(base);
		return true;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		Reg areg = (Reg)restore(nodes, stack, regs, addr, NO_CONST), 
			breg = (Reg)restore(nodes, stack, regs, base, NO_CONST), offset;
		coerce(nodes, stack, regs, areg, ActType.Kind.INT);
		coerce(nodes, stack, regs, breg, ActType.Kind.BP);
		offset = reserve(nodes, stack, regs, 0);
		regs.markNotToSpill(areg); regs.markNotToSpill(breg); regs.markNotToSpill(offset);
		loadLenMetaToScratch(nodes, breg, true);
		boundsCheck(nodes, stack, regs, offset, areg, breg, Reg.scratch, size);
		regs.markUsed(areg);
		
		if (userDest != null) {
			Reg udreg = reserve(nodes, stack, regs, (size == 1 ? NO_EDI : 0));
			nodes.add(new Asm("ld<" + size + ",2>", udreg, offset + "(" + breg + ")"));
			regs.set(udreg, userDest, ActType.freshInt());
			coerce(nodes, regs, udreg);
		}
		if (baseDest != null) {
			Reg bdreg = reserve(nodes, stack, regs, 0);
			loadLenMetaToScratch(nodes, breg, false);
			alignCheck(nodes, stack, regs, offset);
			useMeta(nodes, true, Reg.scratch);
			nodes.add(new Asm("ld<" + size + ",0>", bdreg, offset + "(" + Reg.scratch + ")"));
			unuseMeta(nodes, regs.types[breg.index].len, Reg.scratch);
			regs.set(bdreg, baseDest, ActType.freshPackedBP());
		}
		regs.markUsed(breg);
		regs.set(offset, null, null);
		regs.timestamps[offset.index] = 0;
	}
	
	static void boundsCheck(List<ActNode> nodes, StackLayout stack, RegFile regs, 
							Reg offset, Reg user, Reg base, Reg len, int size)
	{
		nodes.add(new Arith(Arith.Op.ID, false, offset, user, null));
		nodes.add(new Arith(Arith.Op.SUB, false, offset, offset, base));
		nodes.add(new Asm("let", offset, "v" + Utilities.uniqueID()));
		nodes.add(new Jump(Jump.Op.GT, false, Jump.abortLabel, new Const(8), offset).coercionConvert(stack, regs));
		nodes.add(new Arith(Arith.Op.SUB, false, len, len, new Const(size)));
		nodes.add(new Jump(Jump.Op.LT, false, Jump.abortLabel, len, offset).coercionConvert(stack, regs));
		nodes.add(new Arith(Arith.Op.ADD, false, len, len, new Const(size)));
	}
	static void alignCheck(List<ActNode> nodes, StackLayout stack, RegFile regs, Reg offset)
	{
		nodes.add(new Arith(Arith.Op.SUB, false, offset, offset, new Const(8)));
		nodes.add(new Jump(Jump.Op.ND, false, Jump.misalignAbortLabel, new Const(4), offset)
						  .coercionConvert(stack, regs));
	}
	static void useMeta(List<ActNode> nodes, boolean asArray, Reg meta)
	{
		nodes.add(new Asm((asArray ? "unguardth" : "unguardfh"), meta));
	}
	static void unuseMeta(List<ActNode> nodes, Expr len, Reg meta)
	{
		nodes.add(new Asm("guardh", meta, "MetaRec[" + len + "]"));
	}
	
	public String toString()
	{
		return ("(load/" + size + " " + (userDest == null ? "*" : userDest) + " " + 
				(baseDest == null ? "*" : baseDest) + " " + addr + " " + base + ")");
	}
}
