package ctalz.comp.act;

import java.util.*;
import ctalz.util.*;

class Manipulator
{
	private Manipulator()
	{
	}
	
	static void optimize(List<BBlock> blocks)
	{
		VarSet allVars = new VarSet();
		for (BBlock b : blocks) {
			for (ActNode n : b.nodes)
				n.registerVars(allVars);
		}
		VarSet tmpVarSet = allVars.copy();
		for (int count = 0; count < 3; ) {
			boolean done = foldConstant(blocks) & foldJumps(blocks);
			connectNodes(blocks);
			analyzeLiveness(blocks, allVars);
			for (BBlock b : blocks) {
				for (int i = 0; i < b.nodes.size(); i++) {
					ActNode n = b.nodes.get(i);
					if (n instanceof LoadUB && !n.hasSideEffects()) {
						if (!n.liveOut.contains(((LoadUB)n).userDest)) ((LoadUB)n).userDest = null;
						if (!n.liveOut.contains(((LoadUB)n).baseDest)) ((LoadUB)n).baseDest = null;
					}
					if (!n.hasSideEffects()) {
						n.defines(true, tmpVarSet);
						if (tmpVarSet.disjoint(n.liveOut)) {
							b.nodes.remove(i--);
							done = false;
						}
						tmpVarSet.clear();
					}
				}
			}
			count = (done ? count + 1 : 0);
		}
		eliminateJumps(blocks);
		eliminateBlocks(blocks);
	}
	
	static void connectNodes(List<BBlock> blocks)
	{
		for (BBlock b : blocks) {
			for (ActNode n : b.nodes) {
				n.succs = new LinkedList<ActNode>();
				n.preds = new LinkedList<ActNode>();
			}
		}
		for (BBlock b : blocks) {
			for (int i = 0; i < b.nodes.size(); i++) {
				ActNode n = b.nodes.get(i);
				n.addSuccs(i + 1 < b.nodes.size() ? b.nodes.get(i + 1) : null);
				for (ActNode p : n.succs)
					p.preds.add(n);
			}
		}
	}
	
	static boolean foldConstant(List<BBlock> blocks)
	{
		boolean done = true;
		for (BBlock b : blocks) {
			Substitution<ActNode> map = new Substitution<ActNode>();
			for (int i = 0; i < b.nodes.size(); i++) {
				if (!b.nodes.get(i).foldConstant(map, (i + 1 < b.nodes.size() ? b.nodes.get(i + 1) : null))) {
					b.nodes.remove(i--);
					done = false;
				}
			}
		}
		return done;
	}
	static ActNode swapDest(ActNode dest, ActNode next)
	{
		if (next instanceof Arith) {
			Arith a = (Arith)next;
			if (a.op == Arith.Op.ID && a.lhs == dest && (a.dest instanceof Var || !(dest instanceof Var))) {
				a.lhs = a.dest; a.dest = dest;
				return a.lhs;
			}
		}
		return dest;
	}
	
	static boolean foldJumps(List<BBlock> blocks)
	{
		boolean done = true;
		for (BBlock b : blocks) {
			for (int i = 0; i < b.nodes.size(); i++) {
				if (b.nodes.get(i) instanceof Jump) {
					Boolean taken = ((Jump)b.nodes.get(i)).isConstantlyTaken();
					if (taken == null)
						continue;
					if (taken) {
						Label dest = ((Jump)b.nodes.get(i)).dest;
						while (i < b.nodes.size())
							b.nodes.remove(i);
						b.nodes.add(new Jump(dest));
					} else
						b.nodes.remove(i--);
					done = false;
				}
			}
		}
		return done;
	}
	
	static void analyzeLiveness(List<BBlock> blocks, VarSet allVars)
	{
		for (BBlock b : blocks) {
			for (ActNode n : b.nodes) {
				n.liveIn = allVars.copy();
				n.liveOut = allVars.copy();
			}
		}
		boolean done;
		do {
			done = true;
			for (int i = blocks.size() - 1; i >= 0; i--) {
				BBlock b = blocks.get(i);
				for (int j = b.nodes.size() - 1; j >= 0; j--) {
					ActNode n = b.nodes.get(j);
					int oldIn = n.liveIn.size(), oldOut = n.liveOut.size();
					for (ActNode s : n.succs)
						n.liveOut.addAll(s.liveIn);
					n.liveIn.addAll(n.liveOut);
					n.defines(false, n.liveIn);
					n.uses(true, n.liveIn);
					if (n.liveIn.size() != oldIn || n.liveOut.size() != oldOut)
						done = false;
				}
			}
		} while (!done);
	}
	
	static void eliminateJumps(List<BBlock> blocks)
	{
		boolean done;
		Substitution<ActNode> map = new Substitution<ActNode>();
		do {
			done = true;
			for (BBlock b : blocks) {
				if (b.nodes.size() != 2)
				continue;
				ActNode j = b.nodes.get(1);
				if (j instanceof Jump && j.isUnconditionalJump() && ((Jump)j).dest != null)
					done &= !map.put(b.nodes.get(0), ((Jump)j).dest);
			}
			for (BBlock b : blocks) {
				for (ActNode n : b.nodes) {
					if (n instanceof Jump)
						((Jump)n).dest = (Label)map.get(((Jump)n).dest);
				}
			}
		} while (!done);
	}
	
	static void eliminateBlocks(List<BBlock> blocks)
	{
		Set<BBlock> reachable = new HashSet<BBlock>();
		Queue<BBlock> queue = new LinkedList<BBlock>();
		connectNodes(blocks);
		reachable.add(blocks.get(0));
		queue.offer(blocks.get(0));
		while (!queue.isEmpty()) {
			BBlock b = queue.poll();
			for (ActNode n : b.nodes) {
				for (ActNode s : n.succs) {
					if (s instanceof Label && reachable.add(((Label)s).bBlock))
						queue.offer(((Label)s).bBlock);
				}
			}
		}
		ListIterator<BBlock> it = blocks.listIterator();
		while (it.hasNext()) {
			if (!reachable.contains(it.next()))
				it.remove();
		}
	}
}
