package ctalz.comp.act;

import java.util.*;
import ctalz.mac.*;
import ctalz.util.*;

public class NativeCall extends ActNode
{
	protected Var userDest;
	private String label;
	private List<ActNode> args;
	private List<ActType> argTypes;
	private StackLayout savedStack;
	private int stackAlign = -1;
	
	public NativeCall(Program prog, Var u, String l, List<ActNode> a, List<ActType> t)
	{
		prog.registerNativeFunction(l);
		userDest = u; label = l; args = a; argTypes = t;
	}
	
	void registerVars(VarSet set)
	{
		set.register(userDest);
		for (ActNode a : args)
			set.register(a);
	}
	VarSet uses(boolean add, VarSet result)
	{
		for (ActNode a : args)
			if (a instanceof Var) { if (add) result.add((Var)a); else result.remove((Var)a); }
		return result;
	}
	VarSet defines(boolean add, VarSet result)
	{
		if (userDest != null) { if (add) result.add(userDest); else result.remove(userDest); }
		return result;
	}
	boolean hasSideEffects()
	{
		return true;
	}
	
	boolean foldConstant(Substitution<ActNode> map, ActNode next)
	{
		map.removeKeyValue(userDest);
		userDest = (Var)Manipulator.swapDest(userDest, next);
		for (int i = 0; i < args.size(); i++)
			args.set(i, map.get(args.get(i)));
		return true;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		spillAll(nodes, stack, regs);
		nodes.add(this);
		for (int i = args.size() - 1; i >= 0; i--)
			Call.push(nodes, stack, regs, args.get(i), argTypes.get(i), false);
		savedStack = stack.saveTypes();
		nodes.add(this);
		stack.extraElems = 0; stack.extraSize = 0;
		label = Program.stubName(label);
		
		regs.clear();
		if (userDest != null)
			regs.set(Reg.userReturn, userDest, userDest.type);
	}
	void prologueConvert(Program prog, List<ActNode> nodes, StackLayout stack)
	{
		if (stackAlign < 0)
			stackAlign = Call.alignStack(nodes, savedStack.completeMissingTypes());
		else {
			nodes.add(this);
			if (userDest != null && userDest.type.kind == ActType.Kind.INT)
				nodes.add(new Asm("let", Reg.userReturn, userDest.type.val));
			Call.freeStack(nodes, savedStack, stackAlign);
		}
	}
	
	public String toString()
	{
		return ("(call/native " + (userDest == null ? "*" : userDest) + " \"" + label + "\"" + 
				(args.isEmpty() ? "" : " " + Utilities.enumerate(args, " ")) + ")");
	}
	String toAsmString()
	{
		return toAsmString(".unsafe_native", "\"\\tcall " + label + "\"", "[]", "[]");
	}
}
