package ctalz.comp.act;

import java.io.*;
import java.util.*;
import ctalz.comp.*;
import ctalz.comp.ast.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Program
{
	private List<Global> globals = new LinkedList<Global>();
	private List<Function> functions = new LinkedList<Function>();
	private Map<String, Global> strings = new LinkedHashMap<String, Global>();
	private Set<String> nativeFunctions = new LinkedHashSet<String>();
	static final String[] initTemplates = new String[16];
	
	static {
		loadInitTemplates();
	}
	
	public Program()
	{
		loadInitTemplates();
		registerNativeFunction("printf");
	}
	
	public void add(Global g)
	{
		globals.add(g);
	}
	public void add(Function f)
	{
		functions.add(f);
	}
	public ActValue add(String s)
	{
		Global g = strings.get(s);
		if (g == null) {
			strings.put(s, g = new Global(s));
			globals.add(g);
		}
		return new ActValue(new Var(ActType.freshInt(), g.name, true, true, true), 
							new Var(ActType.freshPackedBP(), g.baseName, true, true, true));
	}
	
	void registerNativeFunction(String name)
	{
		nativeFunctions.add(name);
	}
	
	public void emitAssembly(PrintStream out)
	{
		if (!globals.isEmpty()) {
			for (Global g : globals)
				g.emitDefinition(this, out);
			if (Function.dumpACT != null)
				Function.dumpACT.println();
			out.println();
		}
		out.print(initTemplates[0].replace("%init", Arch.encodeLabelName("_.init"))
								  .replace("%hash", Main.uniqueHash()));
		for (Global g : globals)
			g.emitInitializer(this, out);
		out.print(initTemplates[1]);
		out.println();
		for (Function f : functions)
			f.emitAssembly(this, out);
		if (Arch.os() == Arch.OSType.MAC) {
			out.print(initTemplates[10]);
			for (String f : nativeFunctions)
				out.print(initTemplates[11].replace("%fun", Arch.encodeLabelName(f)).replace("%stub", stubName(f)));
		}
		out.print(initTemplates[Arch.os() == Arch.OSType.MAC ? 12 : Arch.os() == Arch.OSType.LINUX ? 13 : 14]
					  .replace("%init", Arch.encodeLabelName(FunDef.initName) + Main.uniqueHash()));
	}
	public static String stubName(String name)
	{
		return (Arch.os() == Arch.OSType.MAC ? "L" + name + "$stub" + Main.uniqueHash() : Arch.encodeLabelName(name));
	}
	
	private static void loadInitTemplates()
	{
		try {
			String[] lines = Utilities.loadResource("/ctalz/comp/init");
			int lineIndex = 0;
			for (int i = 0; i < initTemplates.length; i++) {
				StringBuilder template = new StringBuilder();
				while (lineIndex < lines.length) {
					String line = lines[lineIndex++];
					if (line.startsWith("%%%%%"))
						break;
					template.append(line).append(Utilities.EOL);
				}
				initTemplates[i] = template.toString();
			}
		} catch (IOException e) {
			ExceptionAccumulator.addFatal(Location.NULL, "missingResource", e);
		}
	}
}
