package ctalz.comp.act;

import java.util.*;
import ctalz.comp.*;
import ctalz.comp.ast.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Prologue extends Label
{
	private boolean isPrologue;
	private Var userRet, baseRet;
	private String selfApplyParam;
	private boolean isSecondTimeConversion;
	private static String initializerFunction;
	
	public Prologue(boolean p, String n, Var u, Var b)
	{
		super((p ? n : "end"), p);
		isPrologue = p; userRet = u; baseRet = b;
		if (initializerFunction == null) {
			initializerFunction = Arch.encodeLabelName(Arch.os() == Arch.OSType.MAC ? 
													   FunDef.initName + Main.uniqueHash() : "main");
		}
	}
	
	void registerVars(VarSet set)
	{
		set.register(userRet); set.register(baseRet);
	}
	
	boolean hasSideEffects()
	{
		return true;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		super.coercionConvert(prog, nodes, stack, regs);
		if (isPrologue) {
			if (name.equals(initializerFunction))
				prog.registerNativeFunction("GC_init");
		} else {
			regs.clear();
			if (userRet != null)
				ActNode.restore(nodes, stack, regs, userRet, Reg.userReturn);
			if (baseRet != null)
				ActNode.restore(nodes, stack, regs, baseRet, Reg.baseReturn);
			for (int i = 0; i < Reg.calleeSaves.length; i++)
				ActNode.restore(nodes, stack, regs, stack.calleeSaveVars[i], Reg.calleeSaves[i]);
			nodes.add(this);
		}
	}
	void prologueConvert(Program prog, List<ActNode> nodes, StackLayout stack)
	{
		if (isPrologue) {
			assert labelType == null;
			RegFile regs = new RegFile(stack);
			labelType = new LabelType(requiredStack.initialPart(), regs);
			selfApplyParam = labelType.calculateApplyParams(nodes, requiredStack.initialPart(), regs);
			nodes.add(this);
			addPrologue(nodes, stack);
		} else {
			if (!isSecondTimeConversion)
				nodes.add(this);
			else
				addEpilogue(nodes, stack);
			isSecondTimeConversion = true;
		}
	}
	void addPrologue(List<ActNode> nodes, StackLayout stack)
	{
		StringBuilder sizes = new StringBuilder(), vars = new StringBuilder();
		int totalSize = 0;
		for (int i = stack.types.size() - 1; i >= stack.initialElems; i--) {
			ActType t = stack.types.get(i);
			sizes.append(',').append(t.sizeof());
			totalSize += t.sizeof();
			vars.append(',').append(!stack.isCalleeSave(stack.vars.get(i)) && t.kind == ActType.Kind.INT ? t.val : "_");
		}
		if (totalSize > 0) {
			sizes.setCharAt(0, '<'); sizes.append('>');
			vars.setCharAt(0, '['); vars.append(']');
			nodes.add(new Asm("salloc" + sizes, "$" + totalSize, vars.toString()));
		}
		
		RegFile regs = new RegFile();
		for (int i = 0; i < Reg.calleeSaves.length; i++)
			regs.set(Reg.calleeSaves[i], stack.calleeSaveVars[i], stack.calleeSaveVars[i].type);
		for (Reg r : Reg.calleeSaves)
			spill(nodes, stack, regs, r);
		regs.clear();
		
		Reg zero = (Reg)ActNode.restore(nodes, stack, regs, new Const(0), NO_CONST);
		for (int i = stack.types.size() - 1; i >= stack.initialElems; i--) {
			if (stack.types.get(i).kind == ActType.Kind.PACKED_BP) {
				regs.contents[zero.index] = stack.vars.get(i); regs.modified[zero.index] = true;
				ActNode.spill(nodes, stack, regs, zero);
			}
		}
	}
	void addEpilogue(List<ActNode> nodes, StackLayout stack)
	{
		int elem = 0, size = 0;
		for (int i = stack.types.size() - 1; i >= stack.initialElems; i--, elem++)
			size += stack.types.get(i).sizeof();
		if (size > 0)
			nodes.add(new Asm("sfree<" + elem + ">", "$" + size));
	}
	
	public String toString()
	{
		return (isPrologue ? 
				"(prologue " + (userRet == null ? "*" : userRet) + " " + (baseRet == null ? "*" : baseRet) + ")" : 
				"(epilogue)");
	}
	String toAsmLabelString()
	{
		if (name.equals(initializerFunction) && Arch.os() != Arch.OSType.LINUX) {
			return Program.initTemplates[15].replace("%name", super.toAsmString())
											.replace("%label", super.labelType().toString())
											.replace("%app", selfApplyParam)
											.replace("%fun", Program.stubName((Arch.os() == Arch.OSType.MAC ? 
																			   "GC_init" : "__main")))
											.replace("%hash", Main.uniqueHash()).trim();
		} else
			return super.toAsmLabelString();
	}
}
