package ctalz.comp.act;

import java.util.*;
import ctalz.mac.*;

class RegFile
{
	final ActNode[] contents;
	final ActType[] types;
	final boolean[] modified;
	final int[] timestamps;
	private int nextTimestamp;
	static final int NOT_TO_SPILL = Integer.MAX_VALUE;
	
	private RegFile(ActNode[] c, ActType[] t, boolean[] m, int[] i)
	{
		contents = c; types = t; modified = m; timestamps = i;
	}
	RegFile()
	{
		this(new ActNode[Reg.instances.length], new ActType[Reg.instances.length], 
			 new boolean[Reg.instances.length], new int[Reg.instances.length]);
	}
	RegFile(StackLayout stack)
	{
		this();
		for (int i = 0; i < Reg.calleeSaves.length; i++)
			set(Reg.calleeSaves[i], stack.calleeSaveVars[i], stack.calleeSaveVars[i].type);
		Arrays.fill(modified, false);
	}
	RegFile copy()
	{
		return new RegFile(null, types.clone(), null, null);
	}
	
	Reg find(ActNode v)
	{
		boolean isConst = (v instanceof Const);
		for (int i = 0; i < contents.length; i++) {
			if (i != Reg.scratch.index && v.equals(contents[i]) && (!isConst || types[i].kind == ActType.Kind.INT))
				return Reg.instances[i];
		}
		return null;
	}
	Reg findEmpty(int flags)
	{
		for (int i = 0; i < contents.length; i++) {
			if (i != Reg.scratch.index && 
				contents[i] == null && (Reg.instances[i].noFlags & flags) == 0 && timestamps[i] != NOT_TO_SPILL)
				return Reg.instances[i];
		}
		return null;
	}
	void remove(ActNode v)
	{
		for (int i = 0; i < contents.length; i++) {
			if (v.equals(contents[i])) {
				contents[i] = null;
				types[i] = null;
				timestamps[i] = 0;
			}
		}
	}
	void clear()
	{
		Arrays.fill(contents, null);
		Arrays.fill(types, null);
		Arrays.fill(modified, false);
	}
	
	void markUsed(Reg r)
	{
		timestamps[r.index] = ++nextTimestamp;
	}
	void markNotToSpill(Reg r)
	{
		timestamps[r.index] = NOT_TO_SPILL;
	}
	Reg leastRecentlyUsed(int flags)
	{
		int t = timestamps[0], r = 0;
		for (int i = 1; i < contents.length; i++) {
			if (i != Reg.scratch.index && timestamps[i] < t && (Reg.instances[i].noFlags & flags) == 0) {
				t = timestamps[i];
				r = i;
			}
		}
		return Reg.instances[r];
	}
	
	void set(Reg reg, ActNode val, ActType type)
	{
		contents[reg.index] = val;
		types[reg.index] = type;
		modified[reg.index] = true;
	}
	ActType set(Reg reg, ActType type)
	{
		types[reg.index] = type;
		modified[reg.index] = true;
		return type;
	}
	ActType setWithoutModifying(Reg reg, ActType type)
	{
		types[reg.index] = type;
		return type;
	}
	
	public String toString()
	{
		StringBuilder result = new StringBuilder();
		int count = 0;
		result.append('[');
		for (int i = 0; i < contents.length; i++) {
			if (count++ > 0)
				result.append(", ");
			result.append(Reg.instances[i].name).append(" = ");
			if (contents[i] == null)
				result.append("-");
			else {
				if (contents[i] instanceof Var)
					result.append(((Var)contents[i]).toShortString());
				else
					result.append(((Const)contents[i]).toShortString());
				result.append(':').append(types[i]);
			}
			result.append(modified[i] ? " (*" : " (").append(timestamps[i]).append(')');
		}
		return result.append(']').toString();
	}
}
