package ctalz.comp.act;

import java.util.*;
import ctalz.util.*;

class StackLayout
{
	String tail;
	List<Var> vars;  // in reverse order
	List<ActType> types;
	Var[] calleeSaveVars;
	int initialElems, extraElems, extraSize;
	
	StackLayout(List<Var> args)
	{
		this("s" + Utilities.uniqueID(), new ArrayList<Var>(args), new ArrayList<ActType>(), null, args.size(), 0, 0);
	}
	StackLayout(String t, List<Var> v, List<ActType> y, Var[] c, int i, int e, int s)
	{
		tail = t; vars = v; types = y; calleeSaveVars = c; initialElems = i; extraElems = e; extraSize = s;
		if (calleeSaveVars == null) {
			calleeSaveVars = new Var[Reg.calleeSaves.length];
			for (int j = 0; j < calleeSaveVars.length; j++)
				calleeSaveVars[j] = new Var(ActType.freshInt(), "$calleeSave");
		}
		if (!vars.isEmpty() && types.isEmpty()) {
			for (Var w : vars)
				types.add(w.type);
		}
	}
	StackLayout saveTypes()
	{
		return new StackLayout(tail, vars, new ArrayList<ActType>(types), calleeSaveVars, 
							   initialElems, extraElems, extraSize);
	}
	StackLayout initialPart()
	{
		return new StackLayout(tail, vars.subList(0, initialElems), types.subList(0, initialElems), calleeSaveVars, 
							   initialElems, 0, 0);
	}
	int allocatedSize()
	{
		int result = 0;
		for (int i = initialElems; i < types.size(); i++)
			result += types.get(i).sizeof();
		return result + extraSize;
	}
	
	StackLayout resetTypes()
	{
		for (int i = 0; i < vars.size(); i++)
			types.set(i, vars.get(i).type);
		return this;
	}
	StackLayout completeMissingTypes()
	{
		for (int i = types.size(); i < vars.size(); i++)
			types.add(vars.get(i).type);
		return this;
	}
	
	int find(Var v)
	{
		for (int i = 0; i < vars.size(); i++) {
			if (vars.get(i).equals(v))
				return i;
		}
		vars.add(v);
		types.add(v.type);
		return vars.size() - 1;
	}
	Triple<Integer, Integer, Integer> locate(Var v)
	{
		int at = -1;
		for (int i = 0; i < vars.size(); i++) {
			if (vars.get(i).equals(v)) {
				at = i;
				break;
			}
		}
		assert at >= 0 : "Variable " + v + " is not on stack";
		int offset = 0;
		for (int j = at + 1; j < vars.size(); j++)
			offset += types.get(j).sizeof();
		return new Triple<Integer, Integer, Integer>(types.get(at).sizeof(), 
													 extraElems + vars.size() - at - 1, extraSize + offset);
	}
	
	boolean isCalleeSave(Var v)
	{
		for (Var w : calleeSaveVars) {
			if (v == w)
				return true;
		}
		return false;
	}
	
	public String toString()
	{
		completeMissingTypes();
		
		StringBuilder result = new StringBuilder();
		result.append('[').append(tail);
		for (int i = 0; i < vars.size(); i++)
			result.append(", ").append(vars.get(i).toShortString()).append(':').append(types.get(i));
		return result.append(']').toString();
	}
}
