package ctalz.comp.act;

import java.util.*;
import ctalz.util.*;

public class StoreUB extends ActNode
{
	final int size;
	ActNode userVal, baseVal, addr, base;
	
	public StoreUB(int s, ActNode u, ActNode b, ActNode a, ActNode c)
	{
		assert s == 4 || b == null;
		size = s; userVal = u; baseVal = b; addr = a; base = c;
	}
	
	void registerVars(VarSet set)
	{
		set.register(userVal); set.register(baseVal); set.register(addr); set.register(base);
	}
	VarSet uses(boolean add, VarSet result)
	{
		if (addr instanceof Var) { if (add) result.add((Var)addr); else result.remove((Var)addr); }
		if (base instanceof Var) { if (add) result.add((Var)base); else result.remove((Var)base); }
		if (userVal instanceof Var) { if (add) result.add((Var)userVal); else result.remove((Var)userVal); }
		if (baseVal instanceof Var) { if (add) result.add((Var)baseVal); else result.remove((Var)baseVal); }
		return result;
	}
	boolean hasSideEffects()
	{
		return true;
	}
	
	boolean foldConstant(Substitution<ActNode> map, ActNode next)
	{
		userVal = map.get(userVal); baseVal = map.get(baseVal);
		addr = map.get(addr); base = map.get(base);
		return true;
	}
	
	void coercionConvert(Program prog, List<ActNode> nodes, StackLayout stack, RegFile regs)
	{
		Reg areg = (Reg)restore(nodes, stack, regs, addr, NO_CONST), 
			breg = (Reg)restore(nodes, stack, regs, base, NO_CONST), offset;
		coerce(nodes, stack, regs, areg, ActType.Kind.INT);
		coerce(nodes, stack, regs, breg, ActType.Kind.BP);
		offset = reserve(nodes, stack, regs, 0);
		regs.markNotToSpill(areg); regs.markNotToSpill(breg); regs.markNotToSpill(offset);
		loadLenMetaToScratch(nodes, breg, true);
		LoadUB.boundsCheck(nodes, stack, regs, offset, areg, breg, Reg.scratch, size);
		regs.markUsed(areg);
		
		if (userVal != null) {
			Operand uvreg = restore(nodes, stack, regs, userVal, (size == 1 ? NO_EDI : 0));
			nodes.add(new Asm("st<" + size + ",2>", uvreg, offset + "(" + breg + ")"));
		}
		if (baseVal != null) {
			Reg bvreg = (Reg)restore(nodes, stack, regs, baseVal, NO_CONST);
			coerce(nodes, stack, regs, bvreg, ActType.Kind.PACKED_BP);
			loadLenMetaToScratch(nodes, breg, false);
			LoadUB.alignCheck(nodes, stack, regs, offset);
			LoadUB.useMeta(nodes, true, Reg.scratch);
			nodes.add(new Asm("st<" + size + ",0>", bvreg, offset + "(" + Reg.scratch + ")"));
			LoadUB.unuseMeta(nodes, regs.types[breg.index].len, Reg.scratch);
		}
		regs.markUsed(breg);
		regs.set(offset, null, null);
		regs.timestamps[offset.index] = 0;
	}
	
	public String toString()
	{
		return ("(store/" + size + " " + (userVal == null ? "*" : userVal) + " " + 
				(baseVal == null ? "*" : baseVal) + " " + addr + " " + base + ")");
	}
}
