package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class CastExpr extends Expr
{
	private SpecDecl to;
	private Expr val;
	private CType toC;
	
	public CastExpr(Location loc, SpecDecl t, Expr v)
	{
		super(loc);
		to = t;
		val = v;
	}
	public CastExpr(Location loc, CType t, Expr v)
	{
		super(loc);
		toC = t;
		val = v;
		type = toC;
	}
	
	int priority()
	{
		return 800;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (toC == null)
			toC = to.constructCType(env);
		val = val.constructEnv(env);
		return typeConvert(ctx, toC);
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		val = val.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		CType t = toC.dequalify();
		if (!t.isNumber())
			return null;
		ConstValue v = val.eval();
		if (v == null)
			return null;
		int iv = (v.isInt ? v.extendedIntValue() : (int)v.doubleValue);
		double dv = (v.isInt ? (double)iv : v.doubleValue);
		if (t instanceof BasicType && ((BasicType)t).isIntegral)
			return new ConstValue(t, iv);
		if (t == BasicType.FLOAT || t == BasicType.DOUBLE)
			return new ConstValue(t, dv);
		assert false;
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		CType from = val.type.dequalify(), to = toC.dequalify();
		Arith.Op arithOp;
		if (to == BasicType.VOID) {
			val.compileAsRvalue(prog, fun);
			return new ActValue(null, null);
		} else if (from.equals(to) || 
			((from.isNumber() || from.isPtr()) && (to.isNumber() || to.isPtr()) && from.sizeof() >= to.sizeof())) {
			ActValue v = val.compileAsRvalue(prog, fun);
			return new ActValue(v.userValue, 
								(to.hasBaseValue() ? (v.baseValue != null ? v.baseValue : Const.NULL_BP) : null));
		} else if ((from.isNumber() || from.isPtr()) && (to.isNumber() || to.isPtr()))
			arithOp = (from.sizeof() == 1 ? Arith.Op.EXTEND1 : Arith.Op.EXTEND2);
		else if ((from.isNumber() || from.isPtr()) && (to.isNumber() && to.isPtr())) {
			reportUnsupportedFeature("floatTypeFeature");
			return val.compileAsRvalue(prog, fun);
		} else {
			ExceptionAccumulator.add(loc, "invalidCast", val.type, toC);
			return val.compileAsRvalue(prog, fun);
		}
		ActValue v = val.compileAsRvalue(prog, fun);
		Var u = new Var(to);
		fun.add(new Arith(arithOp, from.isUnsigned(), new VarRef(u), v.userValue, null));
		return new ActValue(u, 
							(to.hasBaseValue() ? (v.baseValue != null ? v.baseValue : Const.NULL_BP) : null));
	}
	
	public String toString()
	{
		String v = val.toString();
		if (val.priority() <= priority())
			v = "(" + v + ")";
		return "(" + (toC != null ? toC : to) + ")" + v;
	}
}
