package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.util.*;

public class CompoundStmt extends Stmt
{
	List<Stmt> stmts;
	
	public CompoundStmt(Location l, List<Stmt> s)
	{
		super(l);
		stmts = s;
	}
	
	public CompoundStmt constructEnv(Env env)
	{
		return constructEnv(true, env);
	}
	CompoundStmt constructEnv(boolean createEnv, Env env)
	{
		if (createEnv)
			env = new Env(env);
		ListIterator<Stmt> it = stmts.listIterator();
		while (it.hasNext())
			it.set(it.next().constructEnv(env));
		return this;
	}
	
	public CompoundStmt transform(List<VarDef> prepend)
	{
		ListIterator<Stmt> it = stmts.listIterator();
		while (it.hasNext())
			it.set(it.next().transform(prepend));
		return this;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		for (Stmt s : stmts)
			s.compile(prog, init, fun);
	}
	
	public void writeTo(Writer out, int indent) throws IOException
	{
		out.append(Utilities.tab(indent - 1)).append('{').append(Utilities.EOL);
		for (Stmt s : stmts) {
			s.writeTo(out, indent);
			out.append(Utilities.EOL);
		}
		out.append(Utilities.tab(indent - 1)).append("}");
	}
}
