package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class ConstExpr extends Expr
{
	final Token val;
	
	public ConstExpr(Location loc, int v)
	{
		super(loc);
		val = new Token(loc, "", Parser.INT_LITERAL, v);
		type = BasicType.INT;
	}
	public ConstExpr(Token v)
	{
		super(v.loc);
		val = v;
	}
	
	boolean isZero()
	{
		return ((val.type == Parser.INT_LITERAL || val.type == Parser.UINT_LITERAL || 
				 val.type == Parser.LONG_LITERAL || val.type == Parser.ULONG_LITERAL) && val.longValue == 0);
	}
	
	int priority()
	{
		return 1000;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		switch (val.type) {
		case Parser.INT_LITERAL:
			type = BasicType.INT;
			break;
		case Parser.UINT_LITERAL:
			type = BasicType.UINT;
			break;
		case Parser.LONG_LITERAL:
			type = BasicType.LONG;
			break;
		case Parser.ULONG_LITERAL:
			type = BasicType.ULONG;
			break;
		case Parser.FLOAT_LITERAL:
		case Parser.DOUBLE_LITERAL:
			reportUnsupportedFeature("constType");
			break;
		case Parser.STR_LITERAL:
			type = new PtrType(BasicType.CHAR);
			break;
		default:
			assert false;
		}
		return this;
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		return this;
	}
	
	public ConstValue eval()
	{
		if (val.type == Parser.INT_LITERAL || val.type == Parser.LONG_LITERAL)
			return new ConstValue(BasicType.INT, (int)val.longValue);
		if (val.type == Parser.UINT_LITERAL || val.type == Parser.ULONG_LITERAL)
			return new ConstValue(BasicType.UINT, (int)val.longValue);
		if (val.type == Parser.FLOAT_LITERAL || val.type == Parser.DOUBLE_LITERAL)
			return new ConstValue(BasicType.DOUBLE, val.doubleValue);
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		if (val.type == Parser.STR_LITERAL)
			return prog.add(val.stringValue);
		return new ActValue(new Const((int)val.longValue), (type.sizeof() >= 4 ? Const.NULL_BP : null));
	}
	
	public String toString()
	{
		return (val.type == Parser.INT_LITERAL ? Long.toString(val.longValue) : 
				val.type == Parser.UINT_LITERAL ? Long.toString(val.longValue) + "U" : 
				val.type == Parser.LONG_LITERAL ? Long.toString(val.longValue) + "L" : 
				val.type == Parser.ULONG_LITERAL ? Long.toString(val.longValue) + "UL" : 
				val.type == Parser.FLOAT_LITERAL ? Double.toString(val.doubleValue) + "f" : 
				val.type == Parser.DOUBLE_LITERAL ? Double.toString(val.doubleValue) : 
				"\"" + Utilities.escape(val.stringValue) + "\"");
	}
}
