package ctalz.comp.ast;

import ctalz.comp.ctype.*;

public class ConstValue
{
	public CType type;
	public int intValue;
	public double doubleValue;
	public boolean isInt;
	
	ConstValue(CType t, int i)
	{
		type = t; intValue = fitSize(t, i); isInt = true;
	}
	ConstValue(CType t, double d)
	{
		type = t; doubleValue = d; isInt = false;
	}
	
	int extendedIntValue()
	{
		if (type == BasicType.CHAR)
			return ((intValue & 0x80) != 0 ? ~0xFF | intValue : intValue);
		if (type == BasicType.SHORT)
			return ((intValue & 0x8000) != 0 ? ~0xFFFF | intValue : intValue);
		return intValue;
	}
	long extendedLongValue()
	{
		if (type == BasicType.UINT || type == BasicType.ULONG)
			return ((long)intValue) & 0xFFFFFFFFL;
		return extendedIntValue();
	}
	
	ConstValue add(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() + v.extendedLongValue()));
	}
	ConstValue sub(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() - v.extendedLongValue()));
	}
	ConstValue mul(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() * v.extendedLongValue()));
	}
	ConstValue div(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		try {
			return new ConstValue(type, fitSize(type, extendedLongValue() / v.extendedLongValue()));
		} catch (Exception e) {
			return null;
		}
	}
	ConstValue mod(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		try {
			return new ConstValue(type, fitSize(type, extendedLongValue() % v.extendedLongValue()));
		} catch (Exception e) {
			return null;
		}
	}
	ConstValue neg()
	{
		if (!isInt)
			return null;
		return new ConstValue(type, fitSize(type, -extendedLongValue()));
	}
	ConstValue bitNot()
	{
		if (!isInt)
			return null;
		return new ConstValue(type, fitSize(type, ~extendedLongValue()));
	}
	ConstValue bitAnd(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() & v.extendedLongValue()));
	}
	ConstValue bitOr(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() | v.extendedLongValue()));
	}
	ConstValue bitXor(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() ^ v.extendedLongValue()));
	}
	ConstValue shiftLeft(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() << v.extendedLongValue()));
	}
	ConstValue shiftRight(ConstValue v)
	{
		if (!isInt || !v.isInt)
			return null;
		return new ConstValue(type, fitSize(type, extendedLongValue() >> v.extendedLongValue()));
	}
	
	private int fitSize(CType type, long value)
	{
		if (type == BasicType.CHAR || type == BasicType.UCHAR)
			return (int)(value & 0xFFL);
		else if (type == BasicType.SHORT || type == BasicType.USHORT)
			return (int)(value & 0xFFFFL);
		return (int)value;
	}
	
	public String toString()
	{
		return isInt ? type + " " + intValue : type + " " + doubleValue;
	}
}
