package ctalz.comp.ast;

import java.util.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class Env
{
	private Map<String, Object> data;
	public final Stack<Label> breakLabels, continueLabels;
	private boolean isFunctionTopLevel;
	private Env parent;
	
	public Env()
	{
		this(null, new Stack<Label>(), new Stack<Label>(), false);
	}
	public Env(Env p)
	{
		this(p, p.breakLabels, p.continueLabels, false);
	}
	public Env(Env p, boolean f)
	{
		this(p, p.breakLabels, p.continueLabels, f);
	}
	private Env(Env p, Stack<Label> b, Stack<Label> c, boolean f)
	{
		data = new HashMap<String, Object>();
		breakLabels = b;
		continueLabels = c;
		isFunctionTopLevel = f;
		parent = p;
	}
	
	public void addLabel(Location loc, String name, Label l)
	{
		findTopLevel().addData(loc, "label", "L" + name, l);
	}
	public void addStruct(Location loc, String name, StructType s)
	{
		addData(loc, "struct", "S" + name, s);
	}
	public void addUnion(Location loc, String name, StructType u)
	{
		addData(loc, "union", "U" + name, u);
	}
	public void addEnum(Location loc, String name, CType e)
	{
		addData(loc, "enum", "E" + name, e);
	}
	public void addEntity(Location loc, String name, Entity e)
	{
		addData(loc, "entity", "V" + name, e);
	}
	
	public Label getLabel(Location loc, String name)
	{
		return (Label)getData(loc, true, "label", "L" + name);
	}
	public StructType getStructOrNull(String name)
	{
		return (StructType)getData(Location.NULL, false, "struct", "S" + name);
	}
	public StructType getUnionOrNull(String name)
	{
		return (StructType)getData(Location.NULL, false, "union", "U" + name);
	}
	public CType getEnum(Location loc, String name)
	{
		return (CType)getData(loc, true, "enum", "E" + name);
	}
	public CType getEnumOrNull(String name)
	{
		return (CType)getData(Location.NULL, false, "enum", "E" + name);
	}
	public Entity getEntity(Location loc, String name)
	{
		return (Entity)getData(loc, true, "entity", "V" + name);
	}
	public Entity getEntityOrNull(String name)
	{
		return (Entity)getData(Location.NULL, false, "entity", "V" + name);
	}
	
	public Env findTopLevel()
	{
		for (Env e = this; e != null; e = e.parent) {
			if (e.isFunctionTopLevel)
				return e;
		}
		return null;
	}
	
	private void addData(Location loc, String cls, String name, Object obj)
	{
		if (data.containsKey(name))
			ExceptionAccumulator.add(loc, "duplicateDefinition", Utilities.localizedString(cls), name.substring(1));
		else
			data.put(name, obj);
	}
	private Object getData(Location loc, boolean reportErr, String cls, String name)
	{
		Object obj = null;
		for (Env e = this; e != null && obj == null; e = e.parent)
			obj = e.data.get(name);
		if (obj == null && reportErr)
			ExceptionAccumulator.add(loc, "undefined", Utilities.localizedString(cls), name.substring(1));
		return obj;
	}
}
