package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public abstract class Expr extends Node
{
	CType type;
	
	public Expr(Location loc)
	{
		super(loc);
	}
	
	public final Expr constructEnv(Env env)
	{
		return constructEnv(Context.NORMAL, false, env);
	}
	abstract Expr constructEnv(Context ctx, boolean pointerTaken, Env env);
	enum Context
	{
		LVALUE, KEEP_ARRAY, NORMAL;
	}
	
	final Expr setType(CType t)
	{
		type = t;
		return this;
	}
	
	final Expr typeConvert(CType t)
	{
		return typeConvert(Context.NORMAL, t);
	}
	final Expr typeConvert(Context ctx, CType t)
	{
		switch (ctx) {
		case KEEP_ARRAY:
		case LVALUE:
			type = t;
			return this;
		case NORMAL:
			t = t.dequalify();
			if (t instanceof ArrayType) {
				Expr e = new UnaryExpr(loc, UnaryExpr.Op.ADDROF, this);
				e.type = new PtrType(((ArrayType)t).base);
				type = t;
				return e;
			} else if (t instanceof FunType)
				type = new PtrType(t);
			else
				type = t;
			return this;
		default:
			assert false;
			return this;
		} 
	}
	
	public abstract Expr transform(List<VarDef> prepend);
	
	public void compile(Program prog, Function init, Function fun)
	{
		compile(prog, fun);
	}
	public void compile(Program prog, Function fun)
	{
		compileAsRvalue(prog, fun);
	}
	ActValue compileAsLvalue(Program prog, Function fun)
	{
		ExceptionAccumulator.add(loc, "notLvalue");
		return new ActValue(new Var(ActType.freshInt()), Const.NULL_BP);
	}
	abstract ActValue compileAsRvalue(Program prog, Function fun);
	static ActValue lvalueToRvalue(Function fun, CType type, ActValue val)
	{
		if (val.userValue instanceof VarRef) {
			return new ActValue(((VarRef)val.userValue).var, 
								(val.baseValue != null ? ((VarRef)val.baseValue).var : null));
		}
		ActNode u = new Var(type), b;
		if (type.hasBaseValue())
			fun.add(new LoadUB(type.sizeof(), u, b = new Var(ActType.freshPackedBP()), val.userValue, val.baseValue));
		else
			fun.add(new LoadUB(type.sizeof(), u, b = null, val.userValue, val.baseValue));
		return new ActValue(u, b);
	}
	
	public abstract ConstValue eval();
	
	abstract int priority();
}
