package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.util.*;

public class ForStmt extends Stmt
{
	private Expr init, cond, step;
	private Stmt body;
	private Label beforeStep = new Label("step"), end = new Label("endfor");
	
	public ForStmt(Location l, Expr i, Expr c, Expr s, Stmt b)
	{
		super(l);
		init = i; cond = c; step = s; body = b;
	}
	
	public Stmt constructEnv(Env env)
	{
		env.breakLabels.push(end);
		env.continueLabels.push(beforeStep);
		if (init != null)
			init = init.constructEnv(env);
		if (cond != null)
			cond = IfStmt.conditionalConvert(cond.constructEnv(env));
		if (step != null)
			step = step.constructEnv(env);
		body = body.constructEnv(env);
		env.breakLabels.pop();
		env.continueLabels.pop();
		return this;
	}
	
	public Stmt transform(List<VarDef> prepend)
	{
		if (init != null)
			init = init.transform(prepend);
		if (cond != null)
			cond = cond.transform(prepend);
		if (step != null)
			step = step.transform(prepend);
		body = body.transform(prepend);
		return this;
	}
	
	public void compile(Program prog, Function initFun, Function fun)
	{
		Label begin = new Label("for");
		if (init != null)
			init.compile(prog, fun);
		fun.add(begin.node);
		if (cond != null)
			((RelExpr)cond).compileAsJump(prog, fun, end.node);
		body.compile(prog, initFun, fun);
		fun.add(beforeStep.node);
		if (step != null)
			step.compile(prog, fun);
		fun.add(new Jump(begin.node));
		fun.add(end.node);
	}
	
	public void writeTo(Writer out, int indent) throws IOException
	{
		out.append(Utilities.tab(indent)).append("for (");
		out.append(init != null ? init.toString() : "").append("; ");
		out.append(cond != null ? cond.toString() : "").append("; ");
		out.append(step != null ? step.toString() : "").append(')').append(Utilities.EOL);
		body.writeTo(out, indent + 1);
	}
}
