package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class IdentExpr extends Expr
{
	String name;
	Entity entity;
	
	public IdentExpr(Location loc, String n)
	{
		super(loc);
		name = n;
	}
	IdentExpr(Location loc, String n, Entity e, CType t)
	{
		super(loc);
		name = n;
		entity = e;
		type = t.dequalify();
	}
	
	int priority()
	{
		return 1000;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		return constructEnv(ctx, pointerTaken, env, false);
	}
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env, boolean isFun)
	{
		entity = env.getEntity(loc, name);
		if (entity == null)
			env.addEntity(loc, name, entity = new Entity(Entity.Kind.VAR, name, BasicType.INT));
		if (pointerTaken && entity.kind == Entity.Kind.VAR)
			entity.isPointerTaken = true;
		if (entity.kind == Entity.Kind.FUN && !isFun)
			entity.isPointerTaken = true;
		return typeConvert(ctx, entity.type);
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		if (entity.transformDeref) {
			return new UnaryExpr(loc, UnaryExpr.Op.DEREF, this.setType(entity.type))
						   .setType(((PtrType)entity.type).base);
		}
		return this;
	}
	
	public ConstValue eval()
	{
		if (entity.kind == Entity.Kind.CONST)
			return new ConstValue(BasicType.INT, entity.val);
		return null;
	}
	
	ActValue compileAsLvalue(Program prog, Function fun)
	{
		return new ActValue(new VarRef(entity.userValue), 
							(entity.baseValue != null ? new VarRef(entity.baseValue) : null));
	}
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		if (entity.kind == Entity.Kind.FUN) {
			if (!entity.fun.hasGeneric())
				ExceptionAccumulator.add(loc, "noGeneric", name);
			return new ActValue(entity.fun.genericUser, entity.fun.genericBase);
		} else if (entity.kind == Entity.Kind.CONST)
			return new ActValue(new Const(entity.val), Const.NULL_BP);
		return new ActValue(entity.userValue, entity.baseValue);
	}
	
	public String toString()
	{
		return name;
	}
}
