package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.util.*;

public class IfStmt extends Stmt
{
	private Expr cond;
	private Stmt then, els;
	
	public IfStmt(Location l, Expr c, Stmt t, Stmt e)
	{
		super(l);
		cond = c; then = t; els = e;
	}
	
	public Stmt constructEnv(Env env)
	{
		cond = conditionalConvert(cond.constructEnv(env));
		then = then.constructEnv(env);
		if (els != null)
			els = els.constructEnv(env);
		return this;
	}
	
	public Stmt transform(List<VarDef> prepend)
	{
		cond = cond.transform(prepend);
		then = then.transform(prepend);
		if (els != null)
			els = els.transform(prepend);
		return this;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		ActNode l1 = new Label("else").node, l2 = new Label("endif").node;
		((RelExpr)cond).compileAsJump(prog, fun, l1);
		then.compile(prog, init, fun);
		if (els != null) {
			fun.add(new Jump(l2));
			fun.add(l1);
			els.compile(prog, init, fun);
			fun.add(l2);
		} else
			fun.add(l1);
	}
	
	static RelExpr conditionalConvert(Expr e)
	{
		if (!e.type.isNumber() && !e.type.isPtr())
			ExceptionAccumulator.add(e.loc, "condType", e.type);
		if (e instanceof RelExpr)
			return (RelExpr)e;
		e = (e.type.dequalify().isPtr() ? e : UnaryExpr.integralPromote(e, "<condition>"));
		return new RelExpr(RelExpr.Op.NE, e, new ConstExpr(e.loc, 0));
	}
	
	public void writeTo(Writer out, int indent) throws IOException
	{
		out.append(Utilities.tab(indent)).append("if (").append(cond.toString()).append(')').append(Utilities.EOL);
		then.writeTo(out, indent + 1);
		if (els != null) {
			out.append(Utilities.EOL).append(Utilities.tab(indent)).append("else").append(Utilities.EOL);
			els.writeTo(out, indent + 1);
		}
	}
}
