package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.util.*;

public class JumpStmt extends Stmt
{
	private Op op;
	private String label;
	private Label target;
	private Env savedEnv;
	
	public enum Op
	{
		GOTO, BREAK, CONTINUE;
	}
	
	public JumpStmt(Location l, Op o)
	{
		super(l);
		op = o;
	}
	public JumpStmt(Location l, Op o, String n)
	{
		super(l);
		op = o; label = n;
	}
	
	public Stmt constructEnv(Env env)
	{
		if (op == Op.GOTO)
			savedEnv = env;
		else {
			if (op == Op.BREAK)
				target = (env.breakLabels.isEmpty() ? null : env.breakLabels.peek());
			else
				target = (env.continueLabels.isEmpty() ? null : env.continueLabels.peek());
			if (target == null)
				ExceptionAccumulator.add(loc, "noLabel");
		}
		return this;
	}
	
	public Stmt transform(List<VarDef> prepend)
	{
		return this;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		if (op == Op.GOTO)
			target = savedEnv.getLabel(loc, label);
		fun.add(new Jump(target.node));
	}
	
	public void writeTo(Writer out, int indent) throws IOException
	{
		out.append(Utilities.tab(indent)).append(op.toString().toLowerCase());
		if (label != null)
			out.append(" ").append(label);
		out.append(";");
	}
}
