package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.util.*;

public class LabelStmt extends Stmt
{
	Kind kind;
	String name;
	Expr expr;
	Stmt body;
	Label label;
	ConstValue exprVal;
	
	public enum Kind
	{
		LABEL, CASE, DEFAULT;
	}
	
	public LabelStmt(Location l, String n, Stmt b)
	{
		super(l);
		kind = Kind.LABEL; name = n; body = b;
	}
	public LabelStmt(Location l, Expr e, Stmt b)
	{
		super(l);
		kind = Kind.CASE; expr = e; body = b;
	}
	public LabelStmt(Location l, Stmt b)
	{
		super(l);
		kind = Kind.DEFAULT; body = b;
	}
	
	public Stmt constructEnv(Env env)
	{
		label = new Label(name);
		if (name != null)
			env.addLabel(loc, name, label);
		if (expr != null) {
			expr = expr.constructEnv(env);
			exprVal = expr.eval();
			if (exprVal == null || !exprVal.type.isIntegral())
				ExceptionAccumulator.add(expr.loc, "shouldBeIntConst");
		}
		body = body.constructEnv(env);
		return this;
	}
	
	public Stmt transform(List<VarDef> prepend)
	{
		if (expr != null)
			expr = expr.transform(prepend);
		if (body != null)
			body = body.transform(prepend);
		return this;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		fun.add(label.node);
		body.compile(prog, init, fun);
	}
	
	public void writeTo(Writer out, int indent) throws IOException
	{
		out.append(Utilities.tab(indent - 1));
		switch (kind) {
		case LABEL: out.append(name).append(':').append(Utilities.EOL); break;
		case CASE: out.append("case ").append(expr.toString()).append(':').append(Utilities.EOL); break;
		case DEFAULT: out.append("default:").append(Utilities.EOL); break;
		}
		body.writeTo(out, indent);
	}
}
