package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class MemberExpr extends Expr
{
	private Op op;
	private Expr val;
	private String name;
	private int offset;
	
	public MemberExpr(Op o, Expr v, String n)
	{
		super(v.loc);
		op = o; val = v; name = n;
	}
	
	enum Op
	{
		SIMPLE("."), 
		DEREF("->");
		
		private String name;
		Op(String n) { name = n; }
	}
	
	int priority()
	{
		return 900;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (op == Op.DEREF) {
			return new MemberExpr(Op.SIMPLE, new UnaryExpr(loc, UnaryExpr.Op.DEREF, val), name)
						   .constructEnv(ctx, pointerTaken, env);
		}
		val = val.constructEnv(env);
		CType t = val.type;
		if (t instanceof StructType) {
			StructType.Member m = ((StructType)t).findMember(name);
			if (m == null)
				ExceptionAccumulator.add(loc, "notMember", name, t);
			else {
				offset = m.offset;
				return typeConvert(ctx, m.type);
			}
		} else
			ExceptionAccumulator.add(loc, "nonStructMem");
		return typeConvert(ctx, BasicType.INT);
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		CType st = val.type;
		Expr ptr = new UnaryExpr(val.loc, UnaryExpr.Op.ADDROF, val.transform(prepend)).setType(new PtrType(st));
		CType cp = new PtrType(BasicType.CHAR);
		ptr = new CastExpr(val.loc, new PtrType(type), 
						   new PtrArithExpr(PtrArithExpr.Op.PTR_PLUS_INT, 
											new CastExpr(val.loc, cp, ptr), 
											new ConstExpr(val.loc, offset)).setType(cp));
		return new UnaryExpr(val.loc, UnaryExpr.Op.DEREF, ptr).setType(type);
	}
	
	public ConstValue eval()
	{
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		assert false;
		return null;
	}
	
	public String toString()
	{
		String v = val.toString();
		if (val.priority() < priority())
			v = "(" + v + ")";
		return v + op.name + name;
	}
}
