package ctalz.comp.ast;

import java.util.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class Params
{
	final List<SpecDecl> newParams;
	final List<String> oldParams;
	final boolean isVarArgs;
	
	public Params(List<SpecDecl> p, boolean v)
	{
		newParams = p; oldParams = null; isVarArgs = v;
	}
	public Params(List<String> p)
	{
		newParams = null; oldParams = p; isVarArgs = false;
	}
	
	public CType constructCType(Env env, CType returnType)
	{
		if (oldParams != null)
			return new FunType(returnType, new LinkedList<Pair<String, CType>>(), true);
		List<Pair<String, CType>> args = new ArrayList<Pair<String, CType>>(newParams.size());
		int onlyVoid = 0;
		env = new Env(env);
		for (SpecDecl s : newParams) {
			String n = s.decl().basemost().name;
			CType t = convertType(s.decl().constructCType(env, Spec.constructCType(env, s.specs)));
			args.add(new Pair<String, CType>(n, t));
			if (n == null && t == BasicType.VOID)
				onlyVoid++;
		}
		if (onlyVoid == 1)
			args.clear();
		return new FunType(returnType, args, isVarArgs);
	}
	private static CType convertType(CType t)
	{
		if (t instanceof ArrayType)
			return new PtrType(((ArrayType)t).base);
		else if (t instanceof FunType)
			return new PtrType(t);
		return t;
	}
	
	public String toString()
	{
		String s = Utilities.enumerate((newParams != null ? newParams : oldParams), ", ");
		if (isVarArgs)
			return (s.equals("") ? "..." : s + ", ...");
		return s;
	}
}
