package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class PtrArithExpr extends Expr
{
	private Op op;
	private Expr lhs, rhs;
	
	public PtrArithExpr(Op o, Expr l, Expr r)
	{
		super(l.loc);
		op = o; lhs = l; rhs = r;
		type = l.type;
	}
	
	enum Op
	{
		PTR_PLUS_INT(BinaryExpr.Op.ADD), 
		PTR_MINUS_INT(BinaryExpr.Op.SUB), 
		PTR_MINUS_PTR(BinaryExpr.Op.SUB);
		
		final BinaryExpr.Op binaryOp;
		Op(BinaryExpr.Op o) { binaryOp = o; }
	}
	
	int priority()
	{
		return op.binaryOp.priority;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (op == Op.PTR_MINUS_PTR) {
			if (!CType.areEquivalentPtrs(lhs.type, rhs.type))
				ExceptionAccumulator.add(lhs.loc, "subIncompatible", lhs.type, rhs.type);
			return typeConvert(ctx, BasicType.PTRDIFF_T);
		} else {
			if (!lhs.type.isPtr()) {
				Expr swap = lhs; lhs = rhs; rhs = swap;
				loc = lhs.loc;
			}
			rhs = UnaryExpr.integralPromote(rhs, op.binaryOp.name.trim());
			if (!rhs.type.isIntegral())
				ExceptionAccumulator.add(rhs.loc, "operandType", op.binaryOp.name.trim(), rhs.type);
			else
				rhs = new CastExpr(rhs.loc, BasicType.INT, rhs);
			return typeConvert(ctx, lhs.type);
		}
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		lhs = lhs.transform(prepend);
		rhs = rhs.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		int s = ((PtrType)lhs.type).base.sizeof();
		if (s <= 0)
			ExceptionAccumulator.add(loc, "sizeNotKnown", ((PtrType)lhs.type).base);
		BinaryExpr e1, e2;
		if (op == Op.PTR_MINUS_PTR) {
			e1 = new BinaryExpr(op.binaryOp, lhs, rhs);
			e2 = new BinaryExpr(BinaryExpr.Op.DIV, e1, new ConstExpr(loc, s));
		} else {
			e1 = new BinaryExpr(BinaryExpr.Op.MUL, rhs, new ConstExpr(loc, s));
			e2 = new BinaryExpr(op.binaryOp, lhs, e1);
		}
		e1.type = e2.type = BasicType.UINT;
		return e2.compileAsRvalue(prog, fun, (op != Op.PTR_MINUS_PTR));
	}
	
	public String toString()
	{
		String l = lhs.toString(), r = rhs.toString();
		if (lhs.priority() < priority())
			l = "(" + l + ")";
		if (rhs.priority() <= priority())
			r = "(" + r + ")";
		return l + op.binaryOp.name + r;
	}
}
