package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.util.*;

public class ShortExpr extends Expr
{
	private Op op;
	private Expr lhs, rhs;
	
	public ShortExpr(Op o, Expr l, Expr r)
	{
		super(l.loc);
		op = o; lhs = l; rhs = r;
	}
	
	enum Op
	{
		OR("||"), 
		AND("&&");
		
		final String name;
		Op(String n) { name = n; }
	}
	
	int priority()
	{
		return 300;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (op == Op.OR) {
			return new TernaryExpr(lhs, new ConstExpr(rhs.loc, 1), 
								   new TernaryExpr(rhs, new ConstExpr(rhs.loc, 1), new ConstExpr(rhs.loc, 0)))
						   .constructEnv(ctx, false, env);
		} else {
			return new TernaryExpr(lhs, new TernaryExpr(rhs, new ConstExpr(rhs.loc, 1), new ConstExpr(rhs.loc, 0)), 
								   new ConstExpr(rhs.loc, 0))
						   .constructEnv(ctx, false, env);
		}
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		assert false;
		return this;
	}
	
	public ConstValue eval()
	{
		assert false;
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		assert false;
		return null;
	}
	
	public String toString()
	{
		String l = lhs.toString(), r = rhs.toString();
		if (lhs.priority() < priority())
			l = "(" + l + ")";
		if (rhs.priority() <= priority())
			r = "(" + r + ")";
		return l + " " + op.name + " " + r;
	}
}
