package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class SizeofExpr extends Expr
{
	private Expr val;
	private SpecDecl ty;
	private CType tyC;
	private int size;
	
	public SizeofExpr(Location loc, Expr v)
	{
		super(loc);
		val = v;
	}
	public SizeofExpr(Location loc, SpecDecl t)
	{
		super(loc);
		ty = t;
	}
	
	int priority()
	{
		return 800;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		if (val != null)
			val = val.constructEnv(Context.KEEP_ARRAY, false, env);
		else
			tyC = ty.constructCType(env);
		size = (val != null ? val.type : tyC).sizeof();
		if (size < 0)
			ExceptionAccumulator.add(loc, "sizeNotKnown", (val != null ? val.type : tyC));
		type = BasicType.SIZE_T;
		return this;
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		return this;
	}
	
	public ConstValue eval()
	{
		return new ConstValue(BasicType.INT, size);
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		return new ActValue(new Const(size), new Const(0));
	}
	
	public String toString()
	{
		return "sizeof(" + (val != null ? val.toString() : ty.toString()) + ")";
	}
}
