package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.comp.ctype.*;
import ctalz.util.*;

public class TernaryExpr extends Expr
{
	private Expr cond, then, els;
	
	public TernaryExpr(Expr c, Expr t, Expr e)
	{
		super(c.loc);
		cond = c; then = t; els = e;
	}
	
	int priority()
	{
		return 200;
	}
	
	Expr constructEnv(Context ctx, boolean pointerTaken, Env env)
	{
		cond = IfStmt.conditionalConvert(cond.constructEnv(env));
		then = then.constructEnv(env);
		els = els.constructEnv(env);
		CType tt = then.type, et = els.type;
		if (tt.isNumber() && et.isNumber()) {
			Pair<Expr, Expr> p = BinaryExpr.arithmeticConvert(then, els, "?:");
			then = p.fst; els = p.snd;
			return typeConvert(ctx, then.type);
		} else if (tt.equals(et))
			return typeConvert(ctx, then.type);
		else if (tt.isPtr() && els instanceof ConstExpr && ((ConstExpr)els).isZero())
			return typeConvert(ctx, then.type);
		else if (et.isPtr() && then instanceof ConstExpr && ((ConstExpr)then).isZero())
			return typeConvert(ctx, els.type);
		else if (tt.isPtr() && et.isPtr() && (tt.isVoidPtr() || et.isVoidPtr()))
			return typeConvert(ctx, new PtrType(BasicType.VOID));
		else {
			ExceptionAccumulator.add(loc, "ternaryIncompatible", tt, et);
			return typeConvert(ctx, BasicType.INT);
		}
	}
	
	public Expr transform(List<VarDef> prepend)
	{
		cond = cond.transform(prepend);
		then = then.transform(prepend);
		els = els.transform(prepend);
		return this;
	}
	
	public ConstValue eval()
	{
		ConstValue c = cond.eval();
		if (c != null)
			return ((c.isInt ? c.intValue != 0 : c.doubleValue != 0) ? then.eval() : els.eval());
		return null;
	}
	
	ActValue compileAsRvalue(Program prog, Function fun)
	{
		Var u = new Var(type);
		ActNode b = (type.hasBaseValue() ? new Var(ActType.freshPackedBP()) : Const.NULL_BP);
		ActNode l1 = new Label("else").node, l2 = new Label("endif").node;
		((RelExpr)cond).compileAsJump(prog, fun, l1);
		ActValue v = then.compileAsRvalue(prog, fun);
		fun.add(new Arith(Arith.Op.ID, false, new VarRef(u), v.userValue, null));
		if (b instanceof Var)
			fun.add(new Arith(Arith.Op.ID, false, new VarRef((Var)b), v.baseValue, null));
		fun.add(new Jump(l2));
		fun.add(l1);
		v = els.compileAsRvalue(prog, fun);
		fun.add(new Arith(Arith.Op.ID, false, new VarRef(u), v.userValue, null));
		if (b instanceof Var)
			fun.add(new Arith(Arith.Op.ID, false, new VarRef((Var)b), v.baseValue, null));
		fun.add(l2);
		return new ActValue(u, b);
	}
	
	public String toString()
	{
		String c = cond.toString(), t = then.toString(), e = els.toString();
		if (cond.priority() <= priority())
			c = "(" + c + ")";
		if (els.priority() < priority())
			e = "(" + e + ")";
		return c + " ? " + t + " : " + e;
	}
}
