package ctalz.comp.ast;

import java.io.*;
import java.util.*;
import ctalz.comp.act.*;
import ctalz.util.*;

public class WhileStmt extends Stmt
{
	private boolean isDoWhile;
	private Expr cond;
	private Stmt body;
	private Label begin = new Label("while"), end = new Label("endwhile");
	
	public WhileStmt(Location l, boolean d, Expr c, Stmt b)
	{
		super(l);
		isDoWhile = d; cond = c; body = b;
	}
	
	public Stmt constructEnv(Env env)
	{
		env.breakLabels.push(end);
		env.continueLabels.push(begin);
		cond = IfStmt.conditionalConvert(cond.constructEnv(env));
		body = body.constructEnv(env);
		env.breakLabels.pop();
		env.continueLabels.pop();
		return this;
	}
	
	public Stmt transform(List<VarDef> prepend)
	{
		cond = cond.transform(prepend);
		body = body.transform(prepend);
		return this;
	}
	
	public void compile(Program prog, Function init, Function fun)
	{
		Label entry = new Label("do");
		if (isDoWhile)
			fun.add(new Jump(entry.node));
		fun.add(begin.node);
		((RelExpr)cond).compileAsJump(prog, fun, end.node);
		if (isDoWhile)
			fun.add(entry.node);
		body.compile(prog, init, fun);
		fun.add(new Jump(begin.node));
		fun.add(end.node);
	}
	
	public void writeTo(Writer out, int indent) throws IOException
	{
		if (isDoWhile) {
			out.append(Utilities.tab(indent)).append("do").append(Utilities.EOL);
			body.writeTo(out, indent + 1);
			out.append(Utilities.EOL);
			out.append(Utilities.tab(indent)).append("while (").append(cond.toString()).append(");");
		} else {
			out.append(Utilities.tab(indent));
			out.append("while (").append(cond.toString()).append(')').append(Utilities.EOL);
			body.writeTo(out, indent + 1);
		}
	}
}
