package ctalz.comp.ctype;

import ctalz.util.*;

public class ArrayType extends CType
{
	public final CType base;
	public final int size;
	
	public ArrayType(Location loc, CType b, int s)
	{
		if (s <= 0) {
			ExceptionAccumulator.add(loc, "negativeArray", s);
			s = 1;
		}
		base = b; size = s;
	}
	public ArrayType(Location loc, CType b)
	{
		base = b; size = -1;
	}
	private ArrayType(CType b, int s)
	{
		base = b; size = s;
	}
	
	public int sizeof()
	{
		return Math.max(-1, base.sizeof()) * size;
	}
	public int alignment()
	{
		return base.alignment();
	}
	
	public CType constType()
	{
		if (constType == null)
			constType = new ArrayType(base.constType(), size);
		return constType;
	}
	
	@Override public boolean equals(Object obj)
	{
		if (!(obj instanceof ArrayType))
			return false;
		return (size == ((ArrayType)obj).size && base.equals(((ArrayType)obj).base));
	}
	@Override public int hashCode()
	{
		return base.hashCode() * size;
	}
	
	protected Pair<String, String> toString(boolean isConst)
	{
		Pair<String, String> b = base.toString(false);
		b.snd = "[" + (size < 0 ? "" : size) + "]" + b.snd;
		return b;
	}
}
