package ctalz.comp.ctype;

import ctalz.util.*;

public class BasicType extends CType
{
	public static final BasicType VOID = new BasicType("void", -1, false, false), 
								  CHAR = new BasicType("char", 1, true, false), 
								  UCHAR = new BasicType("unsigned char", 1, true, true), 
								  SHORT = new BasicType("short", 2, true, false), 
								  USHORT = new BasicType("unsigned short", 2, true, true), 
								  INT = new BasicType("int", 4, true, false), 
								  UINT = new BasicType("unsigned int", 4, true, true), 
								  LONG = new BasicType("long", 4, true, false), 
								  ULONG = new BasicType("unsigned long", 4, true, true), 
								  FLOAT = new BasicType("float", 4, false, false), 
								  DOUBLE = new BasicType("double", 8, false, false);
	public static final BasicType SIZE_T = UINT, PTRDIFF_T = UINT;
	
	public final String name;
	public final int size;
	public final boolean isIntegral, isUnsigned;
	
	private BasicType(String n, int s, boolean i, boolean u)
	{
		name = n; size = s; isIntegral = i; isUnsigned = u;
	}
	
	public int sizeof()
	{
		return size;
	}
	public int alignment()
	{
		return Math.max(size, 0);
	}
	
	public BasicType unsignedType()
	{
		if (this == CHAR) return UCHAR;
		if (this == SHORT) return USHORT;
		if (this == INT) return UINT;
		if (this == LONG) return ULONG;
		return null;
	}
	
	protected Pair<String, String> toString(boolean isConst)
	{
		return new Pair<String, String>(" " + (isConst ? name + " const " : name) + " ", "");
	}
}
