package ctalz.comp.ctype;

import ctalz.util.*;

public abstract class CType
{
	CType constType;
	
	public CType constType()
	{
		if (constType == null)
			constType = new ConstType(this);
		return constType;
	}
	public final CType dequalify()
	{
		return (this instanceof ConstType ? ((ConstType)this).base : this);
	}
	
	public final boolean isConst()
	{
		return (this instanceof ConstType);
	}
	public final boolean isNumber()
	{
		return (this instanceof BasicType && this != BasicType.VOID);
	}
	public final boolean isIntegral()
	{
		return (this instanceof BasicType && ((BasicType)this).isIntegral);
	}
	public final boolean isUnsigned()
	{
		return (this instanceof BasicType && ((BasicType)this).isUnsigned);
	}
	public final boolean isPtr()
	{
		return (this instanceof PtrType);
	}
	public final boolean isVoidPtr()
	{
		return (this instanceof PtrType && ((PtrType)this).base.dequalify() == BasicType.VOID);
	}
	
	public static boolean areEquivalentPtrs(CType t1, CType t2)
	{
		if (!(t1 instanceof PtrType) || !(t2 instanceof PtrType))
			return false;
		return ((PtrType)t1).base.dequalify().equals(((PtrType)t2).base.dequalify());
	}
	
	public abstract int sizeof();
	public abstract int alignment();
	
	public final boolean hasBaseValue()
	{
		CType t = dequalify();
		return ((t.isIntegral() && t.sizeof() == 4) || t.isPtr());
	}
	
	public final String toString()
	{
		Pair<String, String> p = toString(false);
		return (p.fst + " " + p.snd).replaceAll("(\\W) +", "$1").replaceAll(" +(\\W)", "$1").replace("#", " ")
									.replaceAll(" +", " ").trim();
	}
	protected abstract Pair<String, String> toString(boolean isConst);
}
