package ctalz.comp.ctype;

import java.util.*;
import ctalz.util.*;

public class FunType extends CType
{
	public final CType ret;
	public final List<Pair<String, CType>> args;
	public final boolean isVarArgs;
	
	public FunType(CType r, List<Pair<String, CType>> a, boolean v)
	{
		ret = r; args = Collections.unmodifiableList(new ArrayList<Pair<String, CType>>(a)); isVarArgs = v;
	}
	
	public int sizeof()
	{
		return -1;
	}
	public int alignment()
	{
		return 1;
	}
	
	@Override public boolean equals(Object obj)
	{
		if (!(obj instanceof FunType))
			return false;
		FunType t = (FunType)obj;
		if (!ret.equals(t.ret) || args.size() != t.args.size() || isVarArgs != t.isVarArgs)
			return false;
		for (int i = 0; i < args.size(); i++) {
			if (!args.get(i).snd.equals(t.args.get(i).snd))
				return false;
		}
		return true;
	}
	@Override public int hashCode()
	{
		int result = ret.hashCode() + (isVarArgs ? 1 : 0);
		for (Pair<String, CType> a : args)
			result = result * 97 + a.snd.hashCode();
		return result;
	}
	
	protected Pair<String, String> toString(boolean isConst)
	{
		StringBuilder arg = new StringBuilder();
		int count = 0;
		for (Pair<String, CType> a : args) {
			if (count++ > 0)
				arg.append(",#");
			arg.append(a.snd.toString());
		}
		if (isVarArgs && count > 0)
			arg.append(",#...");
		Pair<String, String> r = ret.toString(false);
		r.snd = "(" + arg + ")" + r.snd;
		return r;
	}
}
