package ctalz.comp.ctype;

import java.util.*;
import ctalz.comp.ast.*;
import ctalz.util.*;

public class StructType extends CType
{
	public final boolean isUnion;
	public final String name;
	private Map<String, Member> members;
	int size = -1, alignment = 1;
	
	private StructType(boolean u, String n)
	{
		isUnion = u; name = (n == null ? "<anonymous>" : n);
	}
	public static StructType construct(Location loc, Env env, boolean isUnion, String name, List<VarDef> mems)
	{
		int offset = 0, padID = 0, maxSize = 0, maxAlign = 1;
		StructType result = new StructType(isUnion, name);
		Map<String, Member> map = new LinkedHashMap<String, Member>();
		if (name != null) {
			StructType s = (isUnion ? env.getUnionOrNull(name) : env.getStructOrNull(name));
			boolean add = true;
			if (s != null) {
				if (s.members == null) {
					add = false;
					result = s;
				} else if (mems == null)
					return s;
			}
			if (add) {
				if (!isUnion)
					env.addStruct(loc, name, result);
				else
					env.addUnion(loc, name, result);
			}
		}
		if (mems != null) {
			for (VarDef d : mems) {
				CType base = Spec.constructCType(env, d.specs);
				for (Decl dc : d.decls) {
					CType type = dc.constructCType(env, base);
					int size = type.sizeof(), align = type.alignment();
					if (size < 0) {
						ExceptionAccumulator.add(dc.loc, "sizeNotKnown", type);
						size = 0;
					}
					if (!isUnion) {
						int pads = 0;
						while (size != 0 && (offset + pads) % align != 0)
							pads++;
						while (pads-- > 0) {
							String n = "$pad" + (++padID);
							map.put(n, new Member(n, offset++, BasicType.CHAR));
						}
					} else
						offset = 0;
					String n = dc.basemost().name;
					if (map.containsKey(n))
						ExceptionAccumulator.add(dc.loc, "dupMem", n);
					else
						map.put(n, new Member(n, offset, type));
					offset += size;
					maxSize = Math.max(maxSize, offset);
					maxAlign = Math.max(maxAlign, align);
				}
			}
			while (maxSize % maxAlign != 0)
				maxSize++;
		}
		result.members = (mems != null ? map : null);
		result.size = maxSize;
		result.alignment = maxAlign;
		return result;
	}
	
	public int sizeof()
	{
		return (members != null ? size : -1);
	}
	public int alignment()
	{
		return alignment;
	}
	
	public Member findMember(String name)
	{
		return (members != null ? members.get(name) : null);
	}
	
	public static class Member
	{
		public final String name;
		public final int offset;
		public final CType type;
		
		public Member(String n, int o, CType t)
		{
			name = n; offset = o; type = t;
		}
		
		public String toString()
		{
			return offset + ": " + type + " " + name;
		}
	}
	
	protected Pair<String, String> toString(boolean isConst)
	{
		return new Pair<String, String>((isUnion ? " union#" : " struct#") + name + (isConst ? " const " : " "), "");
	}
}
