package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class ApplyInst extends Inst
{
	private IndexOperand addr;
	private List<TypeOperand> types;
	private List<IndexOperand> indices;
	private Operand label;
	
	public ApplyInst(Location loc, IndexOperand a, List<TypeOperand> t, List<IndexOperand> i)
	{
		super(loc);
		addr = a; indices = i; types = t;
	}
	public ApplyInst(Location loc, Operand l)
	{
		super(loc);
		label = l;
	}
	
	public static List<TypeOperand> defaultTypes(Location loc)
	{
		List<TypeOperand> ts = new ArrayList<TypeOperand>(Register.numberOfGPRs + 1);
		for (int i = 0; i <= Register.numberOfGPRs; i++)
			ts.add(new TypeOperand(loc, i));
		return ts;
	}
	public static List<IndexOperand> defaultIndices(Location loc)
	{
		return new ArrayList<IndexOperand>(0);
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		if (types == null) {
			// recover
			assertOperandKind(label, Operand.Kind.LABEL);
			Pair<Expr, BasicType> ht = tc.state.lookupHeap(label.loc, assertType(tc, label, IntType.class).value());
			ht.snd = tc.typeOfLabeledMemory(label.loc, label.name());
			return;
		}
		
		// apply
		Pair<Expr, BasicType> ht = tc.state.lookupHeap(addr.loc, addr.index(tc));
		CodeType code = assertType(addr.loc, tc, ht.snd, CodeType.class);
		List<Type> ts = TypeOperand.types(tc, types);
		List<TypeVar> typeParams = code.types();
		Map<TypeVar, Type> typeMap = new HashMap<TypeVar, Type>();
		if (ts.size() != typeParams.size())
			throw new InvalidSubstitutionException(loc, simplify(tc, code), (Type)null, ts.size(), typeParams.size());
		for (int i = 0; i < ts.size(); i++) {
			if (ts.get(i) != null)
				typeMap.put(typeParams.get(i), ts.get(i));
		}
		Pair<TypeVar, Type> err = Type.isValidSubstitution(tc.state.solver, typeMap);
		if (err != null)
			throw new InvalidSubstitutionException(loc, simplify(tc, code), err.fst, simplify(tc, err.snd));
		
		List<Expr> is = IndexOperand.indices(tc, indices);
		List<VarExpr> indexParams = code.indices();
		Map<VarExpr, Expr> indexMap = new HashMap<VarExpr, Expr>();
		if (is.size() != indexParams.size())
			throw new InvalidSubstitutionException(loc, simplify(tc, code), (Expr)null, is.size(), indexParams.size());
		for (int i = 0; i < is.size(); i++) {
			if (is.get(i) != null)
				indexMap.put(indexParams.get(i), is.get(i));
		}
		
		code = (CodeType)code.substitute(typeMap, indexMap);
		ht.snd = code;
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
	}
	
	public String toString()
	{
		if (types == null)
			return "recover " + label;
		else
			return "apply " + addr + ", " + parametersToString(types) + ", " + parametersToString(indices);
	}
	public static String parametersToString(List<?> params)
	{
		StringBuilder result = new StringBuilder("[");
		int count = 0;
		for (Object p : params) {
			if (count++ > 0)
				result.append(", ");
			if (p == null)
				result.append('*');
			else
				result.append(p);
		}
		return result.append(']').toString();
	}
}
