package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class CallInst extends Inst
{
	private Operand label;
	private List<TypeOperand> types;
	private List<IndexOperand> indices;
	
	public CallInst(Location loc, Operand l)
	{
		super(loc);
		label = l;
	}
	public CallInst(Location loc, List<TypeOperand> t, List<IndexOperand> i)
	{
		super(loc);
		indices = i; types = t;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		Operand tmp = Operand.newReg(loc, tc.state.tmpRegID);
		if (label != null) {
			// call
			if (!insts.hasNext())
				throw new TypeCheckException(loc, "missingLabelAfter");
			Inst next = insts.next();
			insts.previous();
			if (!(next instanceof LabelInst))
				throw new TypeCheckException(loc, "missingLabelAfter");
			VarExpr val = TypeChecker.freshIndexVar();
			Type packAs = new ExType(loc, Collections.singletonList(val), 
				new HeapType(Collections.singletonList(new UnmodifiablePair<Expr, BasicType>
					(val, tc.state.lookupHeap(loc, 
						((IntType)tc.typeOfLabel(loc, ((LabelInst)next).name())).value()).snd))), 
							ConstProp.TRUE, new IntType(val));
			new ArithInst(loc, ArithInst.Op.MOV, tmp, 
						  Operand.newLabel(loc, ((LabelInst)next).name())).typecheck(tc, insts);
			new PackInst(loc, tmp, new TypeOperand(loc, packAs), 
						 Collections.singletonList(new IndexOperand(loc, tc.state.tmpRegID))).typecheck(tc, insts);
			new PushInst(loc, PushInst.Op.PUSH, Arch.wordSize, tmp).typecheck(tc, insts);
			new JmpInst(loc, JmpInst.Cond.ALWAYS, label).typecheck(tc, insts);
		} else {
			// ret
			new PushInst(loc, PushInst.Op.POP, Arch.wordSize, tmp).typecheck(tc, insts);
			new PackInst(loc, tmp, Collections.singletonList(TypeChecker.freshIndexVar())).typecheck(tc, insts);
			new ApplyInst(loc, new IndexOperand(loc, tc.state.tmpRegID), types, indices).typecheck(tc, insts);
			new JmpInst(loc, JmpInst.Cond.ALWAYS, Operand.newReg(loc, tc.state.tmpRegID)).typecheck(tc, insts);
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		if (label != null)
			out.println("\tcall " + (label.kind() == Operand.Kind.REG ? "*" + label : label.toString(true)));
		else
			out.println("\tret");
	}
	
	public String toString()
	{
		if (label != null)
			return "call " + label;
		else
			return "ret " + ApplyInst.parametersToString(types) + ", " + ApplyInst.parametersToString(indices);
	}
}
