package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.type.*;
import ctalz.tc.checker.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class DataInst extends Inst
{
	private Kind kind;
	private int intValue;
	private byte[] stringValue;
	
	public static enum Kind
	{
		INT, STRING;
	}
	
	public DataInst(Location l, int v)
	{
		super(l);
		kind = Kind.INT; intValue = v;
	}
	public DataInst(Location l, byte[] v)
	{
		super(l);
		kind = Kind.STRING; stringValue = v;
	}
	
	public Kind kind()
	{
		return kind;
	}
	public int intValue()
	{
		return intValue;
	}
	public int size()
	{
		if (kind == Kind.INT)
			return Arch.wordSize;
		else
			return stringValue.length;
	}
	public Type type()
	{
		return new IntType(ConstExpr.valueOf(intValue));
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		if (tc.isReachable())
			throw new TypeCheckException(loc, "dataReachable");
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		out.println("\t" + toString());
	}
	
	public String toString()
	{
		switch (kind) {
		case INT:
			return ".long " + intValue;
		case STRING:
			return ".ascii \"" + Utilities.escape(stringValue) + "\"";
		default:
			assert false;
			return null;
		}
	}
}
