package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class GuardInst extends Inst
{
	private Operand addr;
	private TypeOperand type;
	private boolean takeThen;
	
	public GuardInst(Location loc, Operand a, TypeOperand t)
	{
		super(loc);
		addr = a;
		type = t;
	}
	public GuardInst(Location loc, Operand a, boolean t)
	{
		super(loc);
		addr = a;
		takeThen = t;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		if (!(addr instanceof IndexOperand)) {
			assertOperandKind(addr, Operand.Kind.REG);
			assertNotSp(addr);
			if (type != null)
				assertTypeReg(type.loc, tc, type.type(tc));
		}
		
		if (addr instanceof IndexOperand) {
			Pair<Expr, BasicType> ht = tc.state.lookupHeap(addr.loc, ((IndexOperand)addr).index(tc));
			ht.snd = process(tc, ht.snd);
		} else
			addr.setType(tc, process(tc, addr.type(tc)));
	}
	private BasicType process(TypeChecker tc, Type target) throws TypeCheckException
	{
		if (type != null) {
			// guard
			GuardType t = assertType(tc, type, GuardType.class);
			Prop cond = t.cond();
			if (tc.state.solver.isValid(cond)) {
				if (!target.isEquivalent(t.then(), tc.state.solver)) {
					if (tc.state.solver.isUnsatisfiable(cond))
						assertType(addr.loc, tc, target, t.els());
					else
						throw new TypeMismatchException(addr.loc, simplify(tc, target), simplify(tc, t.then()));
				}
			} else if (tc.state.solver.isUnsatisfiable(cond))
				assertType(addr.loc, tc, target, t.els());
			else
				throw new TypeCheckException(loc, "cannotProveGuard", t.cond());
			return t;
		} else {
			// unguard
			GuardType t = assertType(addr.loc, tc, target, GuardType.class);
			Prop cond = t.cond();
			if (takeThen && tc.state.solver.isValid(cond))
				return t.then();
			else if (!takeThen && tc.state.solver.isUnsatisfiable(cond))
				return t.els();
			else
				throw new TypeCheckException(loc, "cannotProveGuard" + (takeThen ? "" :  "Neg"), t.cond());
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
	}
	
	public String toString()
	{
		String suffix = (addr instanceof IndexOperand ? "h" : "r");
		if (type != null)
			return "guard" + suffix + " " + addr + ", " + type;
		else
			return "unguard" + (takeThen ? "t" : "f") + suffix + " " + addr;
	}
}
