package ctalz.tc.inst;

import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class IndexOperand extends Operand
{
	private Expr index;
	private int regID;
	private String labelName;
	
	public IndexOperand(Location l, Expr i)
	{
		super(l, Operand.Kind.META);
		index = i;
	}
	public IndexOperand(Location l, int r)
	{
		super(l, Operand.Kind.META);
		regID = r;
	}
	public IndexOperand(Location l, String n)
	{
		super(l, Operand.Kind.META);
		labelName = n;
	}
	
	public Expr index(TypeChecker tc) throws TypeCheckException
	{
		if (index != null)
			return index;
		Type t = (labelName != null ? tc.typeOfLabel(loc, labelName) : tc.state.reg[regID]);
		if (!(t instanceof IntType)) {
			throw new TypeCheckException(loc, "typeMismatchForIndex", 
										 this, 
										 t.simplify(tc.state.solver).toString(), 
										 Utilities.capitalize(IntType.typeFormStatic()));
		}
		return ((IntType)t).value();
	}
	public static List<Expr> indices(TypeChecker tc, List<IndexOperand> ops) throws TypeCheckException
	{
		List<Expr> result = new ArrayList<Expr>(ops.size());
		for (IndexOperand op : ops)
			result.add(op == null ? null : op.index(tc));
		return result;
	}
	
	public String toString()
	{
		if (index != null)
			return index.toString();
		else if (labelName != null)
			return "addrof(" + labelName + ")";
		else
			return "indexof(" + State.registerNameForID(regID) + ")";
	}
}
