package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public abstract class Inst
{
	protected Location loc;
	
	public Inst(Location l)
	{
		loc = l;
	}
	
	public Location loc()
	{
		return loc;
	}
	
	public abstract void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException;
	
	protected static void assertOperandKind(Operand op, Operand.Kind... expected) throws TypeCheckException
	{
		for (Operand.Kind k : expected) {
			if (op.kind() == k)
				return;
		}
		Arrays.sort(expected);
		throw new TypeCheckException(op.loc, "invalidOperand", 
									 Utilities.enumerate(expected, Utilities.localizedString("operandOr")));
	}
	protected static void assertNotSp(Operand op) throws TypeCheckException
	{
		if (op.kind() == Operand.Kind.REG && op.value() == Register.spID)
			throw new TypeCheckException(op.loc, "invalidOperandSp");
	}
	protected static void assertType(Location loc, TypeChecker tc, Type t1, Type t2) throws TypeCheckException
	{
		if (!t1.isEquivalent(t2, tc.state.solver))
			throw new TypeMismatchException(loc, simplify(tc, t1), simplify(tc, t2));
	}
	@SuppressWarnings("unchecked")
	public static <T extends Type> T assertType(Location loc, TypeChecker tc, 
												Type t, Class<T> type) throws TypeCheckException
	{
		if (!t.getClass().equals(type)) {
			String expected = null;
			try {
				expected = Utilities.capitalize((String)type.getMethod("typeFormStatic").invoke(null));
			} catch (Exception e) {
			}
			throw new TypeMismatchException(loc, simplify(tc, t), expected);
		}
		return (T)t;
	}
	protected static <T extends Type> T assertType(TypeChecker tc, Operand op, 
												   Class<T> type) throws TypeCheckException
	{
		return assertType(op.loc, tc, op.type(tc), type);
	}
	public static void assertTypeNotLinear(Location loc, TypeChecker tc, Type t) throws TypeCheckException
	{
		if (t instanceof LinearType)
			throw new TypeMismatchException(loc, simplify(tc, t), Utilities.localizedString("nonLinearType"));
	}
	public static void assertTypeReg(Location loc, TypeChecker tc, Type t) throws TypeCheckException
	{
		if (!t.isRegisterKind()) {
			throw new TypeCheckException(loc, 
										 new InvalidKindException(loc, "reg", t.simplify(tc.state.solver)).message());
		}
	}
	
	protected static String simplify(TypeChecker tc, Type t)
	{
		if (t == null)
			return Utilities.localizedString("unknown");
		else
			return t.simplify(tc.state.solver).toString();
	}
	protected static String simplify(TypeChecker tc, Expr e)
	{
		if (e == null)
			return Utilities.localizedString("unknown");
		else
			return tc.state.solver.simplify(e).toString();
	}
	protected static String simplify(TypeChecker tc, Prop p)
	{
		if (p == null)
			return Utilities.localizedString("unknown");
		else
			return tc.state.solver.simplify(p).toString();
	}
	
	public abstract void writeNative(TypeChecker tc, PrintStream out);
	
	public abstract String toString();
}
