package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class LabelInst extends Inst
{
	private String name;
	private BasicType type;
	private Location typeLoc;
	private VarExpr addr;
	private List<Inst> typeProof;
	private Location proofLoc;
	
	public LabelInst(Location l, String n, BasicType t, Location tl, List<Inst> p, Location pl)
	{
		super(l);
		name = n;
		type = t;
		typeLoc = tl;
		addr = new VarExpr("addrof(" + name + ")");
		typeProof = p;
		proofLoc = pl;
		if (Main.isReservedLabelName(name))
			ExceptionAccumulator.add(l, "reservedLabel");
		if (!(type instanceof CodeType) && (!(type instanceof TupleType) || ((TupleType)type).length() != 1))
			ExceptionAccumulator.add(loc, "invalidLabelType", type);
	}
	
	public String name()
	{
		return name;
	}
	public BasicType type()
	{
		return type;
	}
	public VarExpr addr()
	{
		return addr;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.suppressUnreachableWarnings = false;
		if (!name.startsWith("."))
			Progress.set(name);
		if (type instanceof CodeType) {
			// Code label
			if (!typeProof.isEmpty())
				throw new TypeCheckException(proofLoc, "proofNotNeeded");
			CodeType t = (CodeType)type;
			if (tc.isReachable())
				tc.state.checkEntailment(loc, t.reg(), t.heap(), t.cnstr());
			tc.initStateTo(t);
		} else {
			// Data label
			TupleType tuple = (TupleType)type;
			if (tc.isReachable())
				throw new TypeCheckException(loc, "dataReachable");
			if (!insts.hasNext())
				throw new TypeCheckException(loc, "missingDataAfter");
			Inst next = insts.next();
			if (!(next instanceof DataInst)) {
				insts.previous();
				throw new TypeCheckException(loc, "missingDataAfter");
			}
			Progress.advance();
			DataInst data = (DataInst)next;
			ArrayType arr = tuple.element(0);
			Expr s = arr.sizeof(tc.state.solver);
			tc.initState();
			tc.state.heap.clear();
			if (!tc.state.solver.isEquivalent(s, ConstExpr.valueOf(data.size())))
				throw new TypeCheckException(typeLoc, "invalidDataSize", simplify(tc, s), data.size());
			if (arr.elementType() != null && data.kind() == DataInst.Kind.STRING)
				throw new TypeCheckException(typeLoc, "invalidDataString");
			List<List<Inst>> ps = new LinkedList<List<Inst>>();
			ps.add(typeProof);
			tc.checkTypeProof(proofLoc, data.type(), tuple, ps);
			tc.state.solver.assume(ConstProp.FALSE);
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		out.println(name + ":");
	}
	
	public String toString()
	{
		return name + ": " + type + ", " + MallocInst.typeProofToString(typeProof);
	}
}
