package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class MallocInst extends Inst
{
	private Operand dest, size;
	private TypeOperand type;
	private List<List<Inst>> typeProof;
	private Location proofLoc;
	
	public MallocInst(Location loc, Operand d, Operand s, TypeOperand t, List<List<Inst>> p, Location pl)
	{
		super(loc);
		dest = d; size = s; type = t; typeProof = p; proofLoc = pl;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		assertOperandKind(dest, Operand.Kind.REG);
		assertNotSp(dest);
		assertOperandKind(size, Operand.Kind.INT, Operand.Kind.REG);
		
		Expr s = assertType(tc, size, IntType.class).value(), s2;
		LinearType t = assertType(tc, type, LinearType.class);
		tc.checkTypeProof(proofLoc, new IntType(ConstExpr.valueOf(0)), t.body(), typeProof);
		s2 = t.body().sizeof(tc.state.solver);
		if (s2 == null || !tc.state.solver.isValid(new RelProp(RelProp.Op.EQ, s, s2)))
			throw new TypeCheckException(size.loc, "mallocSize", simplify(tc, s), simplify(tc, s2));
		dest.setType(tc, t);
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		String label = Main.uniqueReservedLabelName();
		out.println("\tpushl %eax");
		out.println("\tpushl %ecx");
		out.println("\tpushl %edx");
		if (Arch.os() == Arch.OSType.MAC) {
			out.println("\tpushl %ebx");
			out.println("\tmov %esp, %ebx");
			out.println("\tandl $~15, %esp");
			out.println("\tsubl $12, %esp");
		}
		out.println("\tpushl " + size.toString());
		out.println("\tcall " + Main.mallocFunctionStubName());
		out.println("\tcmpl $0, %eax");
		out.println("\tjne " + label);
		new HaltInst(loc, Operand.newInt(loc, Main.failureExitCode())).writeNative(tc, out);
		out.println(label + ":");
		out.println("\tpopl %edx");
		if (Arch.os() == Arch.OSType.MAC) {
			out.println("\tmov %ebx, %esp");
			out.println("\tpopl %ebx");
		}
		out.println("\tpopl %edx");
		out.println("\tpopl %ecx");
		if (!dest.toString().equals("%eax")) {
			out.println("\tmovl %eax, " + dest);
			out.println("\tpopl %eax");
		}
	}
	
	public String toString()
	{
		StringBuilder proofs = new StringBuilder();
		int count = 0;
		for (List<Inst> p : typeProof) {
			if (count++ > 0)
				proofs.append(", ");
			proofs.append(typeProofToString(p));
		}
		return "malloc " + dest + ", " + size + ", " + type + ", [" + proofs + "]";
	}
	public static String typeProofToString(List<Inst> proof)
	{
		StringBuilder result = new StringBuilder("[");
		int count = 0;
		for (Inst i : proof) {
			if (count++ > 0)
				result.append(' ');
			result.append(i.toString()).append(';');
		}
		return result.append(']').toString();
	}
}
