package ctalz.tc.inst;

import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class Operand
{
	protected Location loc;
	private Kind kind;
	private int value, offset;
	private String name;
	
	public static enum Kind
	{
		INT, LABEL, REG, DISP_INT, DISP_REG, META;
		
		public String toString()
		{
			return Utilities.localizedString("operand" + Utilities.capitalize(name().toLowerCase()));
		}
	}
	
	protected Operand(Location l, Kind k)
	{
		loc = l;
		kind = k;
	}
	
	public static Operand newInt(Location l, int i)
	{
		Operand result = new Operand(l, Kind.INT);
		result.value = i;
		return result;
	}
	public static Operand newLabel(Location l, String n)
	{
		Operand result = new Operand(l, Kind.LABEL);
		result.name = n;
		return result;
	}
	public static Operand newReg(Location l, int r)
	{
		Operand result = new Operand(l, Kind.REG);
		result.value = r;
		return result;
	}
	public static Operand newDispInt(Location l, int r, int o)
	{
		Operand result = new Operand(l, Kind.DISP_INT);
		result.value = r;
		result.offset = o;
		return result;
	}
	public static Operand newDispReg(Location l, int r, int o)
	{
		Operand result = new Operand(l, Kind.DISP_REG);
		result.value = r;
		result.offset = o;
		return result;
	}
	
	public Kind kind()
	{
		return kind;
	}
	
	public int value()
	{
		return value;
	}
	public int offset()
	{
		return offset;
	}
	public String name()
	{
		return name;
	}
	public Type type(TypeChecker tc) throws TypeCheckException
	{
		switch (kind) {
		case INT:
			return new IntType(ConstExpr.valueOf(value));
		case LABEL:
			return tc.typeOfLabel(loc, name);
		case REG:
		case DISP_INT:
		case DISP_REG:
			return tc.state.reg[value];
		default:
			assert false;
			return null;
		}
	}
	public void setType(TypeChecker tc, Type t)
	{
		switch (kind) {
		case REG:
		case DISP_INT:
		case DISP_REG:
			tc.state.reg[value] = t;
			break;
		default:
			assert false;
		}
	}
	public Type offsetType(TypeChecker tc)
	{
		switch (kind) {
		case DISP_INT:
			return new IntType(ConstExpr.valueOf(offset));
		case DISP_REG:
			return tc.state.reg[offset];
		case LABEL:
			return new IntType(ConstExpr.valueOf(0));
		default:
			assert false;
			return null;
		}
	}
	
	public String toString()
	{
		return toString(false, Arch.wordSize);
	}
	public String toString(int size)
	{
		return toString(false, size);
	}
	public String toString(boolean suppressDollarBeforeLabel)
	{
		return toString(suppressDollarBeforeLabel, Arch.wordSize);
	}
	public String toString(boolean suppressDollarBeforeLabel, int size)
	{
		switch (kind) {
		case INT:
			return "$" + value;
		case LABEL:
			return (suppressDollarBeforeLabel ? name : "$" + name);
		case REG:
			return State.registerNameForID(value, size);
		case DISP_INT:
			return offset + "(" + State.registerNameForID(value) + ")";
		case DISP_REG:
			return "(" + State.registerNameForID(value) + "," + State.registerNameForID(offset) + ")";
		default:
			assert false;
			return null;
		}
	}
}
