package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class PackInst extends Inst
{
	private Operand addr;
	private TypeOperand type;
	private List<IndexOperand> packAs;
	private List<VarExpr> unpackAs;
	
	public PackInst(Location loc, Operand a, TypeOperand t, List<IndexOperand> p)
	{
		super(loc);
		addr = a; type = t; packAs = p;
	}
	public PackInst(Location loc, Operand a, List<VarExpr> p)
	{
		super(loc);
		addr = a; unpackAs = p;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		if (!(addr instanceof IndexOperand)) {
			assertOperandKind(addr, Operand.Kind.REG);
			assertNotSp(addr);
			if (type != null)
				assertTypeReg(type.loc, tc, type.type(tc));
		}
		
		if (addr instanceof IndexOperand) {
			Pair<Expr, BasicType> ht = tc.state.lookupHeap(addr.loc, ((IndexOperand)addr).index(tc));
			ht.snd = process(tc, ht.snd);
		} else
			addr.setType(tc, process(tc, addr.type(tc)));
	}
	private BasicType process(TypeChecker tc, Type target) throws TypeCheckException
	{
		if (type != null) {
			// pack
			ExType t = assertType(tc, type, ExType.class), ts;
			List<Expr> is = IndexOperand.indices(tc, packAs);
			List<VarExpr> indexParams = t.vars();
			Map<VarExpr, Expr> indexMap = new HashMap<VarExpr, Expr>();
			if (is.size() != indexParams.size())
				throw new InvalidSubstitutionException(loc, simplify(tc, t), (Expr)null, is.size(), indexParams.size());
			for (int i = 0; i < is.size(); i++)
				indexMap.put(indexParams.get(i), is.get(i));
			ts = (ExType)t.substitute(new HashMap<TypeVar, Type>(), indexMap);
			assertType(addr.loc, tc, target, ts.body());
			tc.state.checkEntailment(loc, ts.heap());
			tc.state.checkEntailment(loc, ts.cnstr());
			return t;
		} else {
			// unpack
			ExType t = assertType(addr.loc, tc, target, ExType.class);
			List<VarExpr> indexParams = t.vars();
			Map<VarExpr, Expr> indexMap = new HashMap<VarExpr, Expr>();
			if (unpackAs.size() != indexParams.size()) {
				throw new InvalidSubstitutionException(loc, simplify(tc, t), 
													   (Expr)null, unpackAs.size(), indexParams.size());
			}
			for (int i = 0; i < unpackAs.size(); i++)
				indexMap.put(indexParams.get(i), unpackAs.get(i));
			t = (ExType)t.substitute(new HashMap<TypeVar, Type>(), indexMap);
			for (UnmodifiablePair<Expr, BasicType> p : t.heap().types())
				tc.state.heap.add(new Pair<Expr, BasicType>(p.fst(), p.snd()));
			tc.state.solver.assume(t.cnstr());
			return t.body();
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
	}
	
	public String toString()
	{
		String suffix = (addr instanceof IndexOperand ? "h" : "r");
		if (type != null)
			return "pack" + suffix + " " + addr + ", " + type + ", [" + Utilities.enumerate(packAs, ",") + "]";
		else
			return "unpack" + suffix + " " + addr + ", [" + Utilities.enumerate(unpackAs, ",") + "]";
	}
}
