package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class PushInst extends Inst
{
	private Op op;
	private int size;
	private Operand value;
	
	public static enum Op
	{
		PUSH, POP;
	}
	
	public PushInst(Location loc, Op o, int s, Operand v)
	{
		super(loc);
		op = o; size = s; value = v;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		if (op == Op.PUSH) {
			new StackInst(loc, Collections.singletonList(size), Operand.newInt(loc, size), 
						  Collections.singletonList(VarExpr.freshInstance())).typecheck(tc, insts);
			new StackInst(loc, StackInst.Op.SST, size, 0, value, Operand.newInt(loc, 0)).typecheck(tc, insts);
		} else {
			new StackInst(loc, StackInst.Op.SLD, size, 0, value, Operand.newInt(loc, 0)).typecheck(tc, insts);
			new StackInst(loc, 1, Operand.newInt(loc, size)).typecheck(tc, insts);
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
		String suffix = Arch.instructionSuffixForSize(size);
		if (size == Arch.wordSize)
			out.println("\t" + op.name().toLowerCase() + suffix + " " + value.toString(size));
		else if (op == Op.PUSH) {
			out.println("\tsubl $" + size + ", %esp");
			out.println("\tmov" + suffix + " " + value.toString(size) + ", 0(%esp)");
		} else {
			out.println("\tmov" + suffix + " 0(%esp), " + value.toString(size));
			out.println("\taddl $" + size + ", %esp");
		}
	}
	
	public String toString()
	{
		return op.name().toLowerCase() + "<" + size + "> " + value;
	}
}
