package ctalz.tc.inst;

import java.io.*;
import java.util.*;
import ctalz.tc.checker.*;
import ctalz.type.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class RollInst extends Inst
{
	private Operand addr;
	private TypeOperand type;
	
	public RollInst(Location loc, Operand a, TypeOperand t)
	{
		super(loc);
		addr = a; type = t;
	}
	public RollInst(Location loc, Operand a)
	{
		super(loc);
		addr = a;
	}
	
	public void typecheck(TypeChecker tc, ListIterator<Inst> insts) throws TypeCheckException
	{
		tc.checkReachability(loc);
		
		if (!(addr instanceof IndexOperand)) {
			assertOperandKind(addr, Operand.Kind.REG);
			assertNotSp(addr);
			if (type != null)
				assertTypeReg(type.loc, tc, type.type(tc));
		}
		
		if (addr instanceof IndexOperand) {
			Pair<Expr, BasicType> ht = tc.state.lookupHeap(addr.loc, ((IndexOperand)addr).index(tc));
			ht.snd = process(tc, ht.snd);
		} else
			addr.setType(tc, process(tc, addr.type(tc)));
	}
	private BasicType process(TypeChecker tc, Type target) throws TypeCheckException
	{
		if (type != null) {
			// roll
			RecType t = assertType(tc, type, RecType.class), ts;
			HashMap<TypeVar, Type> typeMap = new HashMap<TypeVar, Type>();
			typeMap.put(t.var(), t.alphaCopy());
			Pair<TypeVar, Type> err = Type.isValidSubstitution(tc.state.solver, typeMap);
			if (err != null)
				throw new InvalidSubstitutionException(loc, simplify(tc, t), err.fst, simplify(tc, err.snd));
			ts = (RecType)t.substitute(typeMap, new HashMap<VarExpr, Expr>());
			assertType(addr.loc, tc, target, ts.body());
			return t;
		} else {
			// unroll
			RecType t = assertType(addr.loc, tc, target, RecType.class);
			HashMap<TypeVar, Type> typeMap = new HashMap<TypeVar, Type>();
			typeMap.put(t.var(), t.alphaCopy());
			Pair<TypeVar, Type> err = Type.isValidSubstitution(tc.state.solver, typeMap);
			if (err != null)
				throw new InvalidSubstitutionException(loc, simplify(tc, t), err.fst, simplify(tc, err.snd));
			t = (RecType)t.substitute(typeMap, new HashMap<VarExpr, Expr>());
			return t.body();
		}
	}
	
	public void writeNative(TypeChecker tc, PrintStream out)
	{
	}
	
	public String toString()
	{
		String suffix = (addr instanceof IndexOperand ? "h" : "r");
		if (type != null)
			return "roll" + suffix + " " + addr + ", " + type;
		else
			return "unroll" + suffix + " " + addr;
	}
}
