package ctalz.tc.parser;

import java.io.*;
import java.util.*;
import ctalz.util.*;
import static ctalz.tc.parser.Parser.*;

%%

%class Lexer
%implements Parser.yyInput
%unicode
%type Token

%{

private String fileName = "<unknown>";
private int lineNumber = 1, charNumber = 1;
private StringBuilder lineText = new StringBuilder();
private List<Byte> string;
private Location stringLoc;
private int blockCommentNestCount;
private Token token;

public void setFileName(String file)
{
	fileName = file;
}

public boolean advance() throws IOException
{
	token = yylex();
	if (token == null)
		return false;
	return true;
}
public int token()
{
	return token.type;
}
public Token value()
{
	return token;
}
public void close()
{
	try {
		for (int n = lineNumber; lineNumber == n && advance(); )
			;
		yyclose();
	} catch (IOException e) {
	}
}

private Location accept(String str)
{
	Location loc = loc();
	charNumber += str.length();
	lineText.append(str);
	return loc;
}
private void advanceLine()
{
	lineNumber++;
	charNumber = 1;
	lineText = new StringBuilder();
}
public Location loc()
{
	return new Location(fileName, lineNumber, charNumber, lineText);
}
public Location locWithoutChar()
{
	return new Location(fileName, lineNumber, lineText);
}

%}

SP = [ \t]
EOL = \n|\r|\r\n
ID = [a-zA-Z_][a-zA-Z0-9_.$]*
DOT_ID = [a-zA-Z_.][a-zA-Z0-9_.$]*

%state STRING, COMMENT, BLOCKCOMMENT

%%

<YYINITIAL> {
  "int"                         { String t = yytext(); return new Token(accept(t), t, INT, t); }
  "array"                       { String t = yytext(); return new Token(accept(t), t, ARRAY, t); }
  "sptr"                        { String t = yytext(); return new Token(accept(t), t, SPTR, t); }
  "rec"                         { String t = yytext(); return new Token(accept(t), t, REC, t); }
  "ex"                          { String t = yytext(); return new Token(accept(t), t, EX, t); }
  "code"                        { String t = yytext(); return new Token(accept(t), t, CODE, t); }
  "{{"                          { String t = yytext(); return new Token(accept(t), t, LL, t); }
  "}}"                          { String t = yytext(); return new Token(accept(t), t, RL, t); }
  "typeof"                      { String t = yytext(); return new Token(accept(t), t, TYPEOF, t); }
  "addrof"                      { String t = yytext(); return new Token(accept(t), t, INDEXOF, t); }
  "indexof"                     { String t = yytext(); return new Token(accept(t), t, INDEXOF, t); }
  "true"                        { String t = yytext(); return new Token(accept(t), t, TRUE, t); }
  "false"                       { String t = yytext(); return new Token(accept(t), t, FALSE, t); }
  "!="                          { String t = yytext(); return new Token(accept(t), t, NE, t); }
  "<="                          { String t = yytext(); return new Token(accept(t), t, LE, t); }
  ">="                          { String t = yytext(); return new Token(accept(t), t, GE, t); }
  "!|"                          { String t = yytext(); return new Token(accept(t), t, ND, t); }
  "and"                         { String t = yytext(); return new Token(accept(t), t, AND, t); }
  "or"                          { String t = yytext(); return new Token(accept(t), t, OR, t); }
  "imp"                         { String t = yytext(); return new Token(accept(t), t, IMP, t); }
  "equiv"                       { String t = yytext(); return new Token(accept(t), t, EQUIV, t); }
  "not"                         { String t = yytext(); return new Token(accept(t), t, NOT, t); }
  
  "add"                         { String t = yytext(); return new Token(accept(t), t, ADD, t); }
  "sub"                         { String t = yytext(); return new Token(accept(t), t, SUB, t); }
  "mul"                         { String t = yytext(); return new Token(accept(t), t, MUL, t); }
  "div"                         { String t = yytext(); return new Token(accept(t), t, DIV, t); }
  "mod"                         { String t = yytext(); return new Token(accept(t), t, MOD, t); }
  "umul"                        { String t = yytext(); return new Token(accept(t), t, UMUL, t); }
  "udiv"                        { String t = yytext(); return new Token(accept(t), t, UDIV, t); }
  "umod"                        { String t = yytext(); return new Token(accept(t), t, UMOD, t); }
  "lsl"                         { String t = yytext(); return new Token(accept(t), t, LSL, t); }
  "lsr"                         { String t = yytext(); return new Token(accept(t), t, LSR, t); }
  "asr"                         { String t = yytext(); return new Token(accept(t), t, ASR, t); }
  "se1"                         { String t = yytext(); return new Token(accept(t), t, SE1, t); }
  "se2"                         { String t = yytext(); return new Token(accept(t), t, SE2, t); }
  "ze1"                         { String t = yytext(); return new Token(accept(t), t, ZE1, t); }
  "ze2"                         { String t = yytext(); return new Token(accept(t), t, ZE2, t); }
  "xor"                         { String t = yytext(); return new Token(accept(t), t, XOR, t); }
  "neg"                         { String t = yytext(); return new Token(accept(t), t, NEG, t); }
  "mov"                         { String t = yytext(); return new Token(accept(t), t, MOV, t); }
  "ld"                          { String t = yytext(); return new Token(accept(t), t, LD, t); }
  "st"                          { String t = yytext(); return new Token(accept(t), t, ST, t); }
  "lst"                         { String t = yytext(); return new Token(accept(t), t, LST, t); }
  "jmp"                         { String t = yytext(); return new Token(accept(t), t, JMP, t); }
  "beq"                         { String t = yytext(); return new Token(accept(t), t, BEQ, t); }
  "bne"                         { String t = yytext(); return new Token(accept(t), t, BNE, t); }
  "blt"                         { String t = yytext(); return new Token(accept(t), t, BLT, t); }
  "ble"                         { String t = yytext(); return new Token(accept(t), t, BLE, t); }
  "bgt"                         { String t = yytext(); return new Token(accept(t), t, BGT, t); }
  "bge"                         { String t = yytext(); return new Token(accept(t), t, BGE, t); }
  "bltu"                        { String t = yytext(); return new Token(accept(t), t, BLTU, t); }
  "bleu"                        { String t = yytext(); return new Token(accept(t), t, BLEU, t); }
  "bgtu"                        { String t = yytext(); return new Token(accept(t), t, BGTU, t); }
  "bgeu"                        { String t = yytext(); return new Token(accept(t), t, BGEU, t); }
  "bdv"                         { String t = yytext(); return new Token(accept(t), t, BDV, t); }
  "bnd"                         { String t = yytext(); return new Token(accept(t), t, BND, t); }
  "fallthru"                    { String t = yytext(); return new Token(accept(t), t, FALL, t); }
  "call"                        { String t = yytext(); return new Token(accept(t), t, CALL, t); }
  "ret"                         { String t = yytext(); return new Token(accept(t), t, RET, t); }
  "malloc"                      { String t = yytext(); return new Token(accept(t), t, MALLOC, t); }
  "commit"                      { String t = yytext(); return new Token(accept(t), t, COMMIT, t); }
  "salloc"                      { String t = yytext(); return new Token(accept(t), t, SALLOC, t); }
  "sfree"                       { String t = yytext(); return new Token(accept(t), t, SFREE, t); }
  "sld"                         { String t = yytext(); return new Token(accept(t), t, SLD, t); }
  "sst"                         { String t = yytext(); return new Token(accept(t), t, SST, t); }
  "push"                        { String t = yytext(); return new Token(accept(t), t, PUSH, t); }
  "pop"                         { String t = yytext(); return new Token(accept(t), t, POP, t); }
  "packr"                       { String t = yytext(); return new Token(accept(t), t, PACK_R, t); }
  "packh"                       { String t = yytext(); return new Token(accept(t), t, PACK_H, t); }
  "unpackr"                     { String t = yytext(); return new Token(accept(t), t, UNPACK_R, t); }
  "unpackh"                     { String t = yytext(); return new Token(accept(t), t, UNPACK_H, t); }
  "rollr"                       { String t = yytext(); return new Token(accept(t), t, ROLL_R, t); }
  "rollh"                       { String t = yytext(); return new Token(accept(t), t, ROLL_H, t); }
  "unrollr"                     { String t = yytext(); return new Token(accept(t), t, UNROLL_R, t); }
  "unrollh"                     { String t = yytext(); return new Token(accept(t), t, UNROLL_H, t); }
  "guardr"                      { String t = yytext(); return new Token(accept(t), t, GUARD_R, t); }
  "guardh"                      { String t = yytext(); return new Token(accept(t), t, GUARD_H, t); }
  "unguardtr"                   { String t = yytext(); return new Token(accept(t), t, UNGUARDT_R, t); }
  "unguardfr"                   { String t = yytext(); return new Token(accept(t), t, UNGUARDF_R, t); }
  "unguardth"                   { String t = yytext(); return new Token(accept(t), t, UNGUARDT_H, t); }
  "unguardfh"                   { String t = yytext(); return new Token(accept(t), t, UNGUARDF_H, t); }
  "apply"                       { String t = yytext(); return new Token(accept(t), t, APPLY, t); }
  "recover"                     { String t = yytext(); return new Token(accept(t), t, RECOVER, t); }
  "halt"                        { String t = yytext(); return new Token(accept(t), t, HALT, t); }
  "let"                         { String t = yytext(); return new Token(accept(t), t, LET, t); }
  "typedef"                     { String t = yytext(); return new Token(accept(t), t, TYPEDEF, t); }
  ".entry"                      { String t = yytext(); return new Token(accept(t), t, ENTRY, t); }
  ".long"                       { String t = yytext(); return new Token(accept(t), t, DATA_LONG, t); }
  ".ascii"                      { String t = yytext(); return new Token(accept(t), t, DATA_STRING, t); }
  ".directive"                  { String t = yytext(); return new Token(accept(t), t, DIRECTIVE, t); }
  ".unsafe_native"              { String t = yytext(); return new Token(accept(t), t, UNSAFE, t); }
  
  {ID}                          { String t = yytext(); return new Token(accept(t), t, ID, t); }
  {DOT_ID}                      { String t = yytext(); return new Token(accept(t), t, DOT_ID, t); }
  "$"? "-"? [0-9]+              {
    String t = yytext(), i = t;
    int tok = INT_LITERAL;
    Location loc = accept(t);
    if (t.charAt(0) == '$') {
      i = i.substring(1);
      tok = DOLLAR_INT_LITERAL;
    }
    try {
      return new Token(loc, t, tok, Integer.parseInt(i));
    } catch (NumberFormatException e) {
      ExceptionAccumulator.add(loc, "integerLiteralOverflow", t);
      return new Token(loc, t, tok, 0);
    }
  }
  "%" {ID}+                     { String t = yytext(); return new Token(accept(t), t, REG_LITERAL, t); }
  "\""                          { stringLoc = accept(yytext()); string = new ArrayList<Byte>(); yybegin(STRING); }
  "#"|"//"                      { accept(yytext()); yybegin(COMMENT); }
  "/*"                          { accept(yytext()); blockCommentNestCount++; yybegin(BLOCKCOMMENT); }
  {SP}+                         { accept(yytext()); }
  ";"                           { return new Token(accept(yytext()), yytext(), EOL); }
  "\\" {EOL}                    { accept("\\"); advanceLine(); }
  {EOL}                         { Location loc = loc(); advanceLine(); return new Token(loc, yytext(), EOL); }
  [\x00-\xFF]                   { String t = yytext(); return new Token(accept(t), t, (int)t.charAt(0)); }
  .                             { String t = yytext(); return new Token(accept(t), t, UNKNOWN); }
  {SP}* "#" {SP}* [0-9]+ {SP}* "\"" [^\"\n\r]* "\"" ({SP} | [0-9])* {EOL} {
    String t = yytext();
    String[] ts = t.substring(0, t.lastIndexOf('"')).trim().split("[ \t]+");
    if (charNumber == 1 && ts.length == 3) {
      try {
        lineNumber = Integer.parseInt(ts[1]) - 1;
        fileName = ts[2].substring(1, ts[2].length());
      } catch (NumberFormatException e) {
      }
    }
    accept(t);
    advanceLine();
  }
}

<STRING> {
  "\\" [0-7] [0-7] [0-7]        {
    String t = yytext();
    int code = Integer.parseInt(t.substring(1), 8);
    if (code > 255)
      ExceptionAccumulator.add(accept(t), "invalidEscape", t);
    else {
      accept(t);
      string.add((byte)code);
    }
  }
  "\\\""                        { accept(yytext()); string.add((byte)'"'); }
  "\\\'"                        { accept(yytext()); string.add((byte)'\''); }
  "\\n"                         { accept(yytext()); string.add((byte)'\n'); }
  "\\r"                         { accept(yytext()); string.add((byte)'\r'); }
  "\\t"                         { accept(yytext()); string.add((byte)'\t'); }
  "\\\\"                        { accept(yytext()); string.add((byte)'\\'); }
  "\\" .                        {
    String t = yytext(); ExceptionAccumulator.add(accept(t), "invalidEscape", t);
  }
  "\""                          {
    accept(yytext());
    byte[] t = new byte[string.size()]; for (int i = 0; i < t.length; i++) t[i] = string.get(i); string = null;
    yybegin(YYINITIAL); return new Token(stringLoc, "<string literal>", STR_LITERAL, t);
  }
  {EOL}                         {
    String t = yytext(); accept(t); advanceLine(); for (byte b : t.getBytes()) string.add(b);
  }
  .                             { String t = yytext(); accept(t); for (byte b : t.getBytes()) string.add(b); }
}

<COMMENT> {
  {EOL}                         {
    Location loc = loc(); advanceLine(); yybegin(YYINITIAL); return new Token(loc, yytext(), EOL);
  }
  .                             { accept(yytext()); }
}

<BLOCKCOMMENT> {
  {EOL}                         { advanceLine(); }
  "/*"                          { accept(yytext()); blockCommentNestCount++; }
  "*/"                          { accept(yytext()); if (--blockCommentNestCount == 0) yybegin(YYINITIAL); }
  .                             { accept(yytext()); }
}
