package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class ArrayType extends BasicType
{
	private BasicType type;
	private Expr size;
	
	public ArrayType(Expr s)
	{
		size = s;
	}
	protected ArrayType(BasicType t, Expr v)
	{
		this(null, t, v);
	}
	public ArrayType(Location loc, BasicType t, Expr s)
	{
		if (!t.isRegisterKind())
			ExceptionAccumulator.add(new InvalidKindException(loc, "reg", t));
		else
			type = t;
		size = s;
	}
	
	public BasicType elementType()
	{
		return type;
	}
	
	public ArrayType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (type == null) {
			Expr newSize = size.substitute(indexMap);
			return (newSize == size ? this : new ArrayType(newSize));
		} else {
			BasicType newType = (BasicType)type.substitute(copy, typeMap, indexMap);
			Expr newSize = size.substitute(indexMap);
			return (newType == type && newSize == size ? this : new ArrayType(newType, newSize));
		}
	}
	public ArrayType simplify(ConstraintSolver cs)
	{
		if (type == null)
			return new ArrayType(cs.simplify(size));
		else
			return new ArrayType(type.simplify(cs), cs.simplify(size));
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof ArrayType))
			return false;
		ArrayType t = (ArrayType)target;
		if (type == null && t.type != null)
			return false;
		if (type != null && (t.type == null || !type.isEquivalent(t.type, cs, typeMap, indexMap)))
			return false;
		return cs.isEquivalent(size.substitute(indexMap), t.size.substitute(indexMap));
	}
	
	public Expr sizeof(ConstraintSolver cs)
	{
		if (cs.isValid(new RelProp(RelProp.Op.GE, size, ConstExpr.valueOf(0))))
			return size;
		else if (cs.isValid(new RelProp(RelProp.Op.LE, size, ConstExpr.valueOf(0))))
			return ConstExpr.valueOf(0);
		return null;
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("arrayForm");
	}
	public String toString()
	{
		if (type == null)
			return "array(" + size + ")";
		else
			return "array(" + type + ", " + size + ")";
	}
}
