package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class CodeType extends BasicType
{
	private List<TypeVar> types;
	private List<VarExpr> indices;
	private RegFileType reg;
	private HeapType heap;
	private Prop cnstr;
	
	protected CodeType(List<TypeVar> t, List<VarExpr> i, RegFileType r, HeapType h, Prop c)
	{
		this(null, t, i, r, h, c);
	}
	public CodeType(Location loc, List<TypeVar> t, List<VarExpr> i, RegFileType r, HeapType h, Prop c)
	{
		types = t; indices = i; reg = r; heap = h; cnstr = c;
	}
	
	public List<TypeVar> types()
	{
		return Collections.unmodifiableList(types);
	}
	public List<VarExpr> indices()
	{
		return Collections.unmodifiableList(indices);
	}
	public RegFileType reg()
	{
		return reg;
	}
	public HeapType heap()
	{
		return heap;
	}
	public Prop cnstr()
	{
		return cnstr;
	}
	
	protected CodeType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (copy) {
			List<TypeVar> newTypes = new ArrayList<TypeVar>(types.size());
			List<VarExpr> newIndices = new ArrayList<VarExpr>(indices.size());
			for (TypeVar v : types) {
				TypeVar vc = v.freshCopy();
				typeMap.put(v, vc);
				newTypes.add(vc);
			}
			for (VarExpr v : indices) {
				VarExpr vc = v.freshCopy();
				indexMap.put(v, vc);
				newIndices.add(vc);
			}
			return new CodeType(newTypes, newIndices, reg.substitute(copy, typeMap, indexMap), 
								heap.substitute(copy, typeMap, indexMap), cnstr.substitute(indexMap));
		} else {
			List<TypeVar> newTypes = new ArrayList<TypeVar>(types);
			List<VarExpr> newIndices = new ArrayList<VarExpr>(indices);
			RegFileType newReg = reg.substitute(copy, typeMap, indexMap);
			HeapType newHeap = heap.substitute(copy, typeMap, indexMap);
			Prop newCnstr = cnstr.substitute(indexMap);
			boolean tsChanged = newTypes.removeAll(typeMap.keySet());
			boolean isChanged = newIndices.removeAll(indexMap.keySet());
			return (!tsChanged && !isChanged && newReg == reg && newHeap == heap && newCnstr == cnstr ? 
					this : new CodeType((tsChanged ? newTypes : types), (isChanged ? newIndices : indices), 
										newReg, newHeap, newCnstr));
		}
	}
	public CodeType simplify(ConstraintSolver cs)
	{
		return new CodeType(types, indices, reg.simplify(cs), heap.simplify(cs), cs.simplify(cnstr));
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof CodeType))
			return false;
		CodeType t = (CodeType)target;
		if (types.size() != t.types.size() || indices.size() != t.indices.size())
			return false;
		Iterator<TypeVar> typesIt = types.iterator();
		for (TypeVar v2 : t.types) {
			TypeVar v1 = typesIt.next();
			if (!v1.equals(v2))
				typeMap.put(v1, v2);
		}
		Iterator<VarExpr> indicesIt = indices.iterator();
		for (VarExpr v2 : t.indices) {
			VarExpr v1 = indicesIt.next();
			if (!v1.equals(v2))
				indexMap.put(v1, v2);
		}
		return (reg.isEquivalent(t.reg, cs, typeMap, indexMap) && 
				heap.isEquivalent(t.heap, cs, typeMap, indexMap) && 
				cs.isEquivalent(cnstr.substitute(indexMap), t.cnstr.substitute(indexMap)));
	}
	
	public boolean isHeapKind()
	{
		return true;
	}
	public Expr sizeof(ConstraintSolver cs)
	{
		return null;
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("codeForm");
	}
	public String toString()
	{
		return ("code[" + Utilities.enumerate(types, ", ") + "][" + Utilities.enumerate(indices, ", ") + "]" + 
				reg + heap + "[" + (cnstr instanceof ConstProp && ((ConstProp)cnstr).value() ? "" : cnstr) + "]");
	}
}
