package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class ExType extends BasicType
{
	private List<VarExpr> vars;
	private HeapType heap;
	private Prop cnstr;
	private BasicType body;
	
	protected ExType(List<VarExpr> v, HeapType h, Prop c, BasicType b)
	{
		this(null, v, h, c, b);
	}
	public ExType(Location loc, List<VarExpr> v, HeapType h, Prop c, BasicType b)
	{
		vars = v; heap = h; cnstr = c; body = b;
		if (!body.isRegisterKind() && !body.isHeapKind()) {
			ExceptionAccumulator.add(new InvalidKindException(loc, "regHeap", body));
			body = new IntType();
		}
	}
	
	public List<VarExpr> vars()
	{
		return Collections.unmodifiableList(vars);
	}
	public HeapType heap()
	{
		return heap;
	}
	public Prop cnstr()
	{
		return cnstr;
	}
	public BasicType body()
	{
		return body;
	}
	
	protected ExType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (copy) {
			List<VarExpr> newVars = new ArrayList<VarExpr>(vars.size());
			for (VarExpr v : vars) {
				VarExpr vc = v.freshCopy();
				indexMap.put(v, vc);
				newVars.add(vc);
			}
			return new ExType(newVars, heap.substitute(copy, typeMap, indexMap), cnstr.substitute(indexMap), 
							  (BasicType)body.substitute(copy, typeMap, indexMap));
		} else {
			HeapType newHeap = heap.substitute(copy, typeMap, indexMap);
			Prop newCnstr = cnstr.substitute(indexMap);
			BasicType newBody = (BasicType)body.substitute(copy, typeMap, indexMap);
			return (newHeap == heap && newCnstr == cnstr && newBody == body ? 
					this : new ExType(vars, newHeap, newCnstr, newBody));
		}
	}
	public ExType simplify(ConstraintSolver cs)
	{
		return new ExType(vars, heap.simplify(cs), cs.simplify(cnstr), body.simplify(cs));
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof ExType))
			return false;
		ExType t = (ExType)target;
		if (vars.size() != t.vars.size())
			return false;
		Iterator<VarExpr> it = vars.iterator();
		for (VarExpr v2 : t.vars) {
			VarExpr v1 = it.next();
			if (!v1.equals(v2))
				indexMap.put(v1, v2);
		}
		return (heap.isEquivalent(t.heap, cs, typeMap, indexMap) && 
				cs.isEquivalent(cnstr.substitute(indexMap), t.cnstr.substitute(indexMap)) && 
				body.isEquivalent(t.body, cs, typeMap, indexMap));
	}
	
	public boolean isRegisterKind()
	{
		return body.isRegisterKind();
	}
	public boolean isHeapKind()
	{
		return body.isHeapKind();
	}
	public Expr sizeof(ConstraintSolver cs)
	{
		Expr result = body.sizeof(cs);
		if (result == null || !cs.isValid(new RelProp(RelProp.Op.NE, result, ConstExpr.valueOf(0))))
			return null;
		return result;
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("exForm");
	}
	public String toString()
	{
		return ("ex[" + Utilities.enumerate(vars, ", ") + "]" + heap + "[" + 
				(cnstr instanceof ConstProp && ((ConstProp)cnstr).value() ? "" : cnstr) + "]:" + body);
	}
}
