package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class GuardType extends BasicType
{
	private Prop cond;
	private BasicType then, els;
	
	protected GuardType(Prop c, BasicType t, BasicType e)
	{
		this(null, c, t, e);
	}
	public GuardType(Location loc, Prop c, BasicType t, BasicType e)
	{
		boolean isHeap;
		cond = c; then = t; els = e;
		isHeap = then.isHeapKind();
		if (!isHeap && !then.isRegisterKind()) {
			ExceptionAccumulator.add(new InvalidKindException(loc, "regHeap", then));
			then = new IntType();
		}
		if ((!isHeap && !els.isRegisterKind()) || (isHeap && !els.isHeapKind())) {
			ExceptionAccumulator.add(new InvalidKindException(loc, (isHeap ? "heap" : "reg"), els));
			els = (isHeap ? TupleType.EMPTY : new IntType());
		}
	}
	
	public Prop cond()
	{
		return cond;
	}
	public BasicType then()
	{
		return then;
	}
	public BasicType els()
	{
		return els;
	}
	
	public GuardType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		Prop newCond = cond.substitute(indexMap);
		BasicType newThen = (BasicType)then.substitute(copy, typeMap, indexMap), 
				  newEls = (BasicType)els.substitute(copy, typeMap, indexMap);
		return (newCond == cond && newThen == then && newEls == els ? this : new GuardType(newCond, newThen, newEls));
	}
	public GuardType simplify(ConstraintSolver cs)
	{
		return new GuardType(cs.simplify(cond), then.simplify(cs), els.simplify(cs));
	}
	
	protected boolean isEquivalent(Type target, 
								  ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof GuardType))
			return false;
		GuardType t = (GuardType)target;
		return (then.isEquivalent(t.then, cs, typeMap, indexMap) && els.isEquivalent(t.els, cs, typeMap, indexMap) && 
				cs.isEquivalent(cond.substitute(indexMap), t.cond.substitute(indexMap)));
	}
	
	public boolean isRegisterKind()
	{
		return then.isRegisterKind();
	}
	public boolean isHeapKind()
	{
		return then.isHeapKind();
	}
	public Expr sizeof(ConstraintSolver cs)
	{
		Expr s1 = then.sizeof(cs), s2 = els.sizeof(cs);
		if (s1 != null && cs.isValid(cond))
			return s1;
		if (s2 != null && cs.isUnsatisfiable(cond))
			return s2;
		return null;
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("guardForm");
	}
	public String toString()
	{
		return "(" + cond + ") ? " + then + " : " + els;
	}
}
