package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class HeapType extends Type
{
	private List<UnmodifiablePair<Expr, BasicType>> types;
	
	public HeapType(List<UnmodifiablePair<Expr, BasicType>> t)
	{
		types = t;
	}
	public HeapType(Location loc, List<Pair<Expr, BasicType>> ts)
	{
		types = new ArrayList<UnmodifiablePair<Expr, BasicType>>(ts.size());
		for (Pair<Expr, BasicType> p : ts) {
			if (!p.snd().isHeapKind()) {
				ExceptionAccumulator.add(loc, "forHeap", p.fst, 
										 new InvalidKindException(loc, "heap", p.snd).message());
			}
			types.add(new UnmodifiablePair<Expr, BasicType>(p.fst, p.snd));
		}
	}
	
	public List<UnmodifiablePair<Expr, BasicType>> types()
	{
		return Collections.unmodifiableList(types);
	}
	
	public HeapType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		boolean changed = false;
		List<UnmodifiablePair<Expr, BasicType>> newTypes = 
			new ArrayList<UnmodifiablePair<Expr, BasicType>>(types.size());
		for (UnmodifiablePair<Expr, BasicType> p : types) {
			Expr newAddr = p.fst().substitute(indexMap);
			BasicType newType = (BasicType)p.snd().substitute(copy, typeMap, indexMap);
			if (newAddr == p.fst() && newType == p.snd())
				newTypes.add(p);
			else {
				newTypes.add(new UnmodifiablePair<Expr, BasicType>(newAddr, newType));
				changed = true;
			}
		}
		return (changed ? new HeapType(newTypes) : this);
	}
	public HeapType simplify(ConstraintSolver cs)
	{
		List<UnmodifiablePair<Expr, BasicType>> newTypes = 
			new ArrayList<UnmodifiablePair<Expr, BasicType>>(types.size());
		for (UnmodifiablePair<Expr, BasicType> p : types)
			newTypes.add(new UnmodifiablePair<Expr, BasicType>(cs.simplify(p.fst()), p.snd().simplify(cs)));
		return new HeapType(newTypes);
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof HeapType))
			return false;
		HeapType t = (HeapType)target;
		ListIterator<UnmodifiablePair<Expr, BasicType>> it1 = types.listIterator(), it2 = t.types.listIterator();
		while (it1.hasNext() && it2.hasNext()) {
			UnmodifiablePair<Expr, BasicType> p1 = it1.next(), p2 = it2.next();
			if (!cs.isEquivalent(p1.fst().substitute(indexMap), p2.fst().substitute(indexMap)) || 
				!p1.snd().isEquivalent(p2.snd(), cs, typeMap, indexMap)) {
				if (p1.snd().isNull(cs)) {
					it1.previous();
					continue;
				}
				if (p2.snd().isNull(cs)) {
					it2.previous();
					continue;
				}
				return false;
			}
		}
		return (!it1.hasNext() && !it2.hasNext());
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("heapForm");
	}
	public String toString()
	{
		StringBuilder result = new StringBuilder("[");
		int count = 0;
		for (UnmodifiablePair<Expr, BasicType> p : types) {
			if (count++ > 0)
				result.append(", ");
			result.append(p.fst()).append(" = ").append(p.snd());
		}
		return result.append(']').toString();
	}
}
