package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class IntType extends BasicType
{
	private Expr value;
	
	public IntType(Expr v)
	{
		value = v;
	}
	public IntType()
	{
		value = VarExpr.freshInstance();
	}
	
	public Expr value()
	{
		return (value == null ? VarExpr.freshInstance() : value);
	}
	
	public IntType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (value == null)
			return this;
		Expr newValue = value.substitute(indexMap);
		return (newValue == value ? this : new IntType(newValue));
	}
	public IntType simplify(ConstraintSolver cs)
	{
		if (value == null)
			return this;
		return new IntType(cs.simplify(value));
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof IntType))
			return false;
		IntType t = (IntType)target;
		return ((value == null && t.value == null) || 
				(value != null && t.value != null && 
				 cs.isEquivalent(value.substitute(indexMap), t.value.substitute(indexMap))));
	}
	
	public boolean isRegisterKind()
	{
		return true;
	}
	public Expr sizeof(ConstraintSolver cs)
	{
		return ConstExpr.valueOf(Arch.wordSize);
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("intForm");
	}
	public String toString()
	{
		return "int(" + value + ")";
	}
}
