package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class RecType extends BasicType
{
	private TypeVar var;
	private BasicType body;
	
	protected RecType(TypeVar v, BasicType t)
	{
		this(null, v, t);
	}
	public RecType(Location loc, TypeVar v, BasicType t)
	{
		boolean isReg;
		var = v; body = t;
		isReg = body.isRegisterKind();
		if (!isReg && !body.isHeapKind()) {
			ExceptionAccumulator.add(new InvalidKindException(loc, "regHeap", body));
			body = new IntType();
		}
		if ((isReg && !var.isRegisterKind()) || (!isReg && !var.isHeapKind())) {
			ExceptionAccumulator.add(new InvalidKindException(loc, (isReg ? "reg" : "heap"), var));
			body = new IntType();
		}
	}
	
	public TypeVar var()
	{
		return var;
	}
	public BasicType body()
	{
		return body;
	}
	
	protected RecType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (copy) {
			TypeVar newVar = var.freshCopy();
			typeMap.put(var, newVar);
			return new RecType(newVar, (BasicType)body.substitute(copy, typeMap, indexMap));
		} else {
			BasicType newBody = (BasicType)body.substitute(copy, typeMap, indexMap);
			return (newBody == body ? this : new RecType(var, newBody));
		}
	}
	public RecType simplify(ConstraintSolver cs)
	{
		return new RecType(var, body.simplify(cs));
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof RecType))
			return false;
		RecType t = (RecType)target;
		if (!var.equals(t.var))
			typeMap.put(var, t.var);
		return body.isEquivalent(t.body, cs, typeMap, indexMap);
	}
	
	public boolean isRegisterKind()
	{
		return body.isRegisterKind();
	}
	public boolean isHeapKind()
	{
		return body.isHeapKind();
	}
	public Expr sizeof(ConstraintSolver cs)
	{
		return body.sizeof(cs);
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("recForm");
	}
	public String toString()
	{
		return "rec[" + var + "]:" + body;
	}
}
