package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class RegFileType extends Type
{
	private Type[] types;
	
	public RegFileType(Location loc, List<Pair<Integer, Type>> t)
	{
		types = new Type[Register.numberOfGPRs + 1];
		for (Pair<Integer, Type> p : t) {
			if (types[p.fst] != null) {
				ExceptionAccumulator.add(loc, "duplicateRegKey", Register.nameForID(p.fst));
				continue;
			}
			if (p.fst != Register.spID) {
				if (!p.snd.isRegisterKind() && !(p.snd instanceof LinearType)) {
					ExceptionAccumulator.add(loc, "forRegister", Register.nameForID(p.fst), 
											 new InvalidKindException(loc, "regLinear", p.snd).message());
					types[p.fst] = new IntType(ConstExpr.valueOf(0));
					continue;
				}
			} else {
				if (!p.snd.isStackKind()) {
					ExceptionAccumulator.add(loc, "forRegister", Register.nameForID(p.fst), 
											 new InvalidKindException(loc, "stack", p.snd).message());
					types[p.fst] = new StackType();
					continue;
				}
			}
			types[p.fst] = p.snd;
		}
		for (int i = 0; i <= Register.numberOfGPRs; i++) {
			if (types[i] == null)
				ExceptionAccumulator.add(loc, "regKeyNotSpecified", Register.nameForID(i));
		}
	}
	public RegFileType(Type[] t)
	{
		assert t.length == Register.numberOfGPRs + 1;
		types = t.clone();
	}
	
	public Type[] types()
	{
		return types.clone();
	}
	public Type typeof(int i)
	{
		return types[i];
	}
	
	public RegFileType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		Type[] newTypes = new Type[types.length];
		boolean changed = false;
		for (int i = 0; i < types.length; i++) {
			Type newType = types[i].substitute(copy, typeMap, indexMap);
			newTypes[i] = newType;
			changed |= (newType != types[i]);
		}
		return (changed ? new RegFileType(newTypes) : this);
	}
	public RegFileType simplify(ConstraintSolver cs)
	{
		Type[] newTypes = new Type[types.length];
		for (int i = 0; i < types.length; i++)
			newTypes[i] = types[i].simplify(cs);
		return new RegFileType(newTypes);
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof RegFileType))
			return false;
		RegFileType t = (RegFileType)target;
		for (int i = 0; i < types.length; i++) {
			if (!types[i].isEquivalent(t.types[i], cs, typeMap, indexMap))
				return false;
		}
		return true;
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("regFileForm");
	}
	public String toString()
	{
		StringBuilder result = new StringBuilder("[");
		for (int i = 0; i <= Register.numberOfGPRs; i++) {
			if (i != 0)
				result.append(", ");
			result.append(Register.nameForID(i)).append(" = ").append(types[i]);
		}
		return result.append("]").toString();
	}
}
