package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class StackType extends Type
{
	private List<BasicType> list;
	private TypeVar tail;
	
	public StackType()
	{
		list = new ArrayList<BasicType>(0);
	}
	private StackType(List<BasicType> l, TypeVar t)
	{
		list = l; tail = t;
	}
	public StackType(Location loc, List<? extends BasicType> ts)
	{
		list = new ArrayList<BasicType>(ts);
		if (!list.isEmpty() && list.get(list.size() - 1) instanceof TypeVar) {
			TypeVar v = (TypeVar)list.get(list.size() - 1);
			if (v.range() == null || v.range() == TypeVar.Range.STACK_TYPE) {
				tail = v;
				tail.setRange(TypeVar.Range.STACK_TYPE);
				list.remove(list.size() - 1);
			}
		}
		for (Iterator<BasicType> it = list.iterator(); it.hasNext(); ) {
			BasicType t = it.next();
			if (!t.isRegisterKind()) {
				ExceptionAccumulator.add(new InvalidKindException(loc, "reg", t));
				it.remove();
			}
		}
	}
	
	public int length()
	{
		return list.size();
	}
	public StackType append(List<BasicType> ts)
	{
		List<BasicType> newList = new ArrayList<BasicType>(ts);
		newList.addAll(list);
		return new StackType(newList, tail);
	}
	public StackType remove(int n)
	{
		List<BasicType> newList = new ArrayList<BasicType>(list);
		while (n-- > 0)
			newList.remove(0);
		return new StackType(newList, tail);
	}
	public BasicType element(int index)
	{
		return list.get(index);
	}
	public StackType updateElement(int index, BasicType t)
	{
		List<BasicType> newList = new ArrayList<BasicType>(list);
		newList.set(index, t);
		return new StackType(newList, tail);
	}
	public Expr offsetOf(ConstraintSolver cs, int index)
	{
		return TupleType.offsetOf(list, cs, index);
	}
	
	public StackType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		List<BasicType> newList = new ArrayList<BasicType>(list.size());
		boolean changed = false;
		for (Type t : list) {
			BasicType newType = (BasicType)t.substitute(copy, typeMap, indexMap);
			newList.add(newType);
			changed |= (newType != t);
		}
		Type newTail = tail.substitute(copy, typeMap, indexMap);
		if (newTail instanceof TypeVar)
			return (!changed && newTail == tail ? this : new StackType((changed ? newList : list), (TypeVar)newTail));
		newList.addAll(((StackType)newTail).list);
		return new StackType(newList, ((StackType)newTail).tail);
	}
	public StackType simplify(ConstraintSolver cs)
	{
		List<BasicType> newList = new ArrayList<BasicType>(list.size());
		for (BasicType t : list)
			newList.add(t.simplify(cs));
		return new StackType(newList, tail);
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof StackType))
			return false;
		StackType t = (StackType)target;
		if (list.size() != t.list.size())
			return false;
		Iterator<BasicType> it = list.iterator();
		for (BasicType v2 : t.list) {
			Type v1 = it.next();
			if (!v1.isEquivalent(v2, cs, typeMap, indexMap))
				return false;
		}
		if ((tail == null) != (t.tail == null))
			return false;
		return (tail == null || tail.isEquivalent(t.tail, cs, typeMap, indexMap));
	}
	
	public boolean isStackKind()
	{
		return true;
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("stackForm");
	}
	public String toString()
	{
		List<BasicType> types = new ArrayList<BasicType>(list);
		if (tail != null)
			types.add(tail);
		return "sptr{" + Utilities.enumerate(types, ", ") + "}";
	}
}
