package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.util.*;

public class TupleType extends BasicType
{
	private List<ArrayType> types;
	public static final TupleType EMPTY = new TupleType(new ArrayList<ArrayType>(0));
	
	protected TupleType(List<ArrayType> ts)
	{
		types = ts;
	}
	public TupleType(Location loc, List<BasicType> ts)
	{
		Iterator<BasicType> it = ts.iterator();
		types = new ArrayList<ArrayType>(ts.size());
		while (it.hasNext()) {
			Type s = it.next();
			if (s instanceof ArrayType)
				types.add((ArrayType)s);
			else if (s.isRegisterKind()) {
				Expr size = ((BasicType)s).sizeof(null);
				assert size != null;
				types.add(new ArrayType((BasicType)s, size));
			} else {
				ExceptionAccumulator.add(new InvalidKindException(loc, "array", s));
				it.remove();
			}
		}
	}
	
	public int length()
	{
		return types.size();
	}
	public List<ArrayType> elements()
	{
		return Collections.unmodifiableList(types);
	}
	public ArrayType element(int index)
	{
		return types.get(index);
	}
	public BasicType updateElement(int index, ArrayType t)
	{
		List<ArrayType> newTypes = new ArrayList<ArrayType>(types);
		newTypes.set(index, t);
		return new TupleType(newTypes);
	}
	public Expr offsetOf(ConstraintSolver cs, int index)
	{
		return offsetOf(types, cs, index);
	}
	protected static <T extends BasicType> Expr offsetOf(List<T> ts, ConstraintSolver cs, int index)
	{
		Iterator<T> it = ts.iterator();
		Expr result = ConstExpr.valueOf(0);
		for (int i = 0; i < index; i++) {
			BasicType t = it.next();
			Expr s = t.sizeof(cs);
			if (s == null)
				return null;
			result = OpExpr.calculate(OpExpr.Op.ADD, result, s);
		}
		return result;
	}
	
	public TupleType substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		List<ArrayType> newTypes = new ArrayList<ArrayType>(types.size());
		boolean changed = false;
		for (ArrayType t : types) {
			ArrayType newType = t.substitute(copy, typeMap, indexMap);
			newTypes.add(newType);
			changed |= (newType != t);
		}
		return (changed ? new TupleType(newTypes) : this);
	}
	public TupleType simplify(ConstraintSolver cs)
	{
		List<ArrayType> newTypes = new ArrayList<ArrayType>(types.size());
		for (ArrayType t : types)
			newTypes.add(t.simplify(cs));
		return new TupleType(newTypes);
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		if (!(target instanceof TupleType))
			return false;
		TupleType t = (TupleType)target;
		if (types.size() != t.types.size())
			return false;
		Iterator<ArrayType> it = types.iterator();
		for (ArrayType v2 : t.types) {
			Type v1 = it.next();
			if (!v1.isEquivalent(v2, cs, typeMap, indexMap))
				return false;
		}
		return true;
	}
	
	public boolean isHeapKind()
	{
		return true;
	}
	public Expr sizeof(ConstraintSolver cs)
	{
		return offsetOf(cs, types.size());
	}
	
	public String typeForm()
	{
		return typeFormStatic();
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("tupleForm");
	}
	public String toString()
	{
		return "{" + Utilities.enumerate(types, ", ") + "}";
	}
}
