package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public abstract class Type
{
	public final Type alphaCopy()
	{
		return substitute(true, new HashMap<TypeVar, Type>(), new HashMap<VarExpr, Expr>());
	}
	public final Type substitute(Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		return substitute(false, typeMap, indexMap);
	}
	public Type substituteAndAlphaCopy(Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		return substitute(true, typeMap, indexMap);
	}
	protected abstract Type substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap);
	public abstract Type simplify(ConstraintSolver cs);
	
	public final boolean isEquivalent(Type target, ConstraintSolver cs)
	{
		return (this == target || isEquivalent(target, cs, new HashMap<TypeVar, Type>(), new HashMap<VarExpr, Expr>()));
	}
	protected abstract boolean isEquivalent(Type target, 
										ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap);
	
	public boolean isRegisterKind()
	{
		return false;
	}
	public boolean isHeapKind()
	{
		return false;
	}
	public boolean isStackKind()
	{
		return false;
	}
	
	public static Pair<TypeVar, Type> isValidSubstitution(ConstraintSolver cs, Map<TypeVar, Type> map)
	{
		for (Map.Entry<TypeVar, Type> entry : map.entrySet()) {
			TypeVar v = entry.getKey();
			boolean ok = true;
			if (v.isRegisterKind()) {
				if (!entry.getValue().isRegisterKind())
					ok = false;
			} else if (v.isHeapKind()) {
				if (!entry.getValue().isHeapKind())
					ok = false;
			} else if (v.range() == TypeVar.Range.STACK_TYPE) {
				if (!entry.getValue().isStackKind())
					ok = false;
			} else
				assert false;
			if (!ok)
				return new Pair<TypeVar, Type>(v, entry.getValue());
		}
		return null;
	}
	
	public abstract String typeForm();
	public abstract String toString();
}
