package ctalz.type;

import java.util.*;
import ctalz.cnstr.*;
import ctalz.mac.*;
import ctalz.util.*;

public class TypeVar extends BasicType
{
	private String name;
	private Range range;
	private Type bind;
	private int id;
	private static boolean showID = (Config.asInt("ShowVariableID", 0) != 0);
	
	public static enum Range
	{
		REGISTER_TYPE, HEAP_TYPE, STACK_TYPE;
		
		public String toString()
		{
			if (this == REGISTER_TYPE)
				return Utilities.localizedString("rangeReg");
			else if (this == HEAP_TYPE)
				return Utilities.localizedString("rangeHeap");
			else
				return Utilities.localizedString("rangeStack");
		}
	}
	
	public TypeVar(String n)
	{
		this(n, null);
	}
	public TypeVar(String n, Range r)
	{
		name = n;
		range = r;
		id = Utilities.uniqueID();
	}
	public TypeVar freshCopy()
	{
		return new TypeVar(name, range);
	}
	
	public String name()
	{
		return name;
	}
	public Range range()
	{
		return range;
	}
	public void setRange(Range r)
	{
		range = r;
	}
	public Type bind()
	{
		return bind;
	}
	public void bind(Type t)
	{
		if (bind == null)
			bind = t;
	}
	
	public Type substitute(boolean copy, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		Type t = typeMap.get(this);
		return (t != null ? t : this);
	}
	public TypeVar simplify(ConstraintSolver cs)
	{
		return this;
	}
	
	protected boolean isEquivalent(Type target, 
								   ConstraintSolver cs, Map<TypeVar, Type> typeMap, Map<VarExpr, Expr> indexMap)
	{
		return (target instanceof TypeVar && 
				substitute(typeMap, indexMap).equals(target.substitute(typeMap, indexMap)));
	}
	
	public boolean isRegisterKind()
	{
		if (range == null)
			range = Range.REGISTER_TYPE;
		return (range == Range.REGISTER_TYPE);
	}
	public boolean isHeapKind()
	{
		if (range == null)
			range = Range.HEAP_TYPE;
		return (range == Range.HEAP_TYPE);
	}
	public Expr sizeof(ConstraintSolver cs)
	{
		return null;
	}
	
	public String typeForm()
	{
		if (range == null)
			return typeFormStatic();
		switch (range) {
		case REGISTER_TYPE:
			return Utilities.localizedString("varRegForm");
		case HEAP_TYPE:
			return Utilities.localizedString("varHeapForm");
		case STACK_TYPE:
			return Utilities.localizedString("varStackForm");
		default:
			assert false;
			return null;
		}
	}
	public static String typeFormStatic()
	{
		return Utilities.localizedString("varForm");
	}
	public String toString()
	{
		return (showID ? name + "#" + id : name);
	}
}
