package ctalz.util;

import java.io.*;
import java.util.*;

public class ExceptionAccumulator
{
	private static List<CTALException> exceptions = new LinkedList<CTALException>();
	private static int numberOfWarnings;
	private static boolean dontReportWarnings;
	public static int maxNumberOfExceptions = Config.asInt("MaxErrorsToReport", 1);
	
	private ExceptionAccumulator()
	{
	}
	
	public static void setDontReportWarnings(boolean value)
	{
		dontReportWarnings = value;
	}
	
	public static void add(CTALException e)
	{
		if (dontReportWarnings && e.isWarning())
			return;
		for (CTALException e2 : exceptions) {
			if (e2.loc().equals(e.loc()) && (e.isWarning() || !e2.isWarning()))
				return;
		}
		exceptions.add(e);
		if (e.isWarning())
			numberOfWarnings++;
		if (e.isFatal() || numberOfExceptions() >= maxNumberOfExceptions)
			throw new AbortError();
	}
	public static void add(Location loc, String format, Object... args)
	{
		add(new CTALException(loc, format, args));
	}
	public static void addWarning(Location loc, String format, Object... args)
	{
		add(new CTALException(loc, format, args).setWarning());
	}
	public static void addFatal(Location loc, String format, Object... args)
	{
		add(new CTALException(loc, format, args).setFatal());
	}
	
	public static boolean isEmpty()
	{
		return exceptions.isEmpty();
	}
	public static int numberOfExceptions()
	{
		return exceptions.size() - numberOfWarnings;
	}
	public static int numberOfWarnings()
	{
		return numberOfWarnings;
	}
	
	public static void printlnTo(PrintStream out)
	{
		int count = 0;
		Collections.sort(exceptions, new Comparator<CTALException>() {
			public int compare(CTALException e1, CTALException e2) { return e1.loc().compareTo(e2.loc()); }
		});
		for (CTALException e : exceptions) {
			if (count++ > 0)
				out.println();
			String loc = e.loc().toString(), context = e.loc().context();
			if (e.isWarning())
				out.println(Utilities.localizedString(loc.equals("") ? "warning" : "warningAt", loc));
			else
				out.println(Utilities.localizedString(loc.equals("") ? "error" : "errorAt", loc));
			if (!context.equals(""))
				out.println(Utilities.indent(context));
			out.println(Utilities.indent(e.message()));
		}
	}
}
