package ctalz.util;

public class Location implements Comparable<Location>
{
	public static final Location NULL = new Location(null);
	
	private String file;
	private int line, chr;
	private StringBuilder text;
	private static final boolean showContext = (Config.asInt("IncludeContextInErrorMessages", 1) != 0);
	
	public Location(String f, int l, int c, StringBuilder t)
	{
		this.file = f;
		this.line = l;
		this.chr = c;
		this.text = (showContext ? t : null);
	}
	public Location(String f, int l, StringBuilder t)
	{
		this.file = f;
		this.line = l;
		this.text = (showContext ? t : null);
	}
	public Location(String f)
	{
		this.file = f;
	}
	
	public String toString()
	{
		if (this == NULL)
			return Utilities.localizedString("locNull");
		else if (line <= 0)
			return Utilities.localizedString("locFile", file);
		else if (chr <= 0)
			return Utilities.localizedString("locFileLine", file, line);
		else
			return Utilities.localizedString("locFileLineChar", file, line, chr);
	}
	
	public String context()
	{
		if (this == NULL || text == null)
			return "";
		String line = text.toString();
		StringBuilder indicator = new StringBuilder();
		int i, length = line.length(), lastSpace = -1;
		boolean inLeadingSpaces = true, indicatorAppended = false;
		for (i = 0; i < length; i++) {
			if (i + 1 >= chr) {
				indicator.append(Utilities.localizedString("indicator"));
				indicatorAppended = true;
				break;
			}
			char chr = line.charAt(i);
			if (chr == ' ' || chr == '\t') {
				if (inLeadingSpaces) {
					lastSpace = i;
					continue;
				}
			} else
				inLeadingSpaces = false;
			indicator.append(chr == '\t' ? Utilities.tab() : chr <= 0xFF ? " " : "  ");
		}
		line = line.substring(lastSpace + 1);
		if (!indicatorAppended)
			indicator.append(Utilities.localizedString("indicator"));
		return line.replace("\t", Utilities.tab()) + Utilities.EOL + indicator;
	}
	
	public int compareTo(Location target)
	{
		long t1 = ((long)line << 32L) + chr, t2 = ((long)target.line << 32L) + target.chr;
		if (file == null) {
			if (target.file == null)
				return 0;
			else
				return 1;
		} else {
			if (target.file == null)
				return -1;
			else {
				int result = file.compareTo(target.file);
				if (result != 0)
					return result;
				return Long.signum(t1 - t2);
			}
		}
	}
	@Override public boolean equals(Object obj)
	{
		if (!(obj instanceof Location))
			return false;
		Location target = (Location)obj;
		return ((file == target.file || 
				 (file != null && target.file != null && file.equals(target.file))) && 
				line == target.line && chr == target.chr);
	}
	@Override public int hashCode()
	{
		return (file != null ? file.hashCode() : 0) + line * 101 + chr * 997;
	}
}
