package ctalz.util;

public class Progress
{
	private static int screenWidth, messageWidth;
	private static boolean begun;
	private static String messagePrefix, message;
	private static int current, maximum, shown;
	
	private Progress()
	{
	}
	
	static {
		String width = Config.asString("ScreenWidth", "50");
		if (width.startsWith("+")) {
			try {
				screenWidth = Math.max(1, Integer.parseInt(width.substring(1)));
			} catch (NumberFormatException e) {
			}
		}
		if (screenWidth <= 0) {
			try {
				screenWidth = Math.max(1, Integer.parseInt(System.getenv("COLUMNS")));
			} catch (Exception e) {
			}
		}
		if (screenWidth <= 0) {
			try {
				screenWidth = Math.max(1, Integer.parseInt(width));
			} catch (NumberFormatException e) {
			}
		}
		if (screenWidth <= 0)
			screenWidth = 50;
		messageWidth = Math.max(-1, Config.asInt("ShowProgress", 0));
		messagePrefix = Config.asString("ProgressMessagePrefix", "");
	}
	
	public static int screenWidth()
	{
		return screenWidth;
	}
	
	public static void begin()
	{
		if (messageWidth != 0 && screenWidth >= messageWidth + 5) {
			begun = true;
			set(0, 1, "");
		}
	}
	public static void end()
	{
		if (begun) {
			System.err.println();
			begun = false;
		}
	}
	
	public static void set(String msg)
	{
		message = messagePrefix + msg;
		shown = -1;
		show();
	}
	public static void set(int cur, int max, String msg)
	{
		current = cur;
		maximum = max;
		set(msg);
	}
	public static void setMaximum(int max)
	{
		maximum = max;
	}
	public static void advance()
	{
		current++;
		show();
	}
	private static void show()
	{
		if (!begun)
			return;
		
		final int barWidth = screenWidth - messageWidth - 2;
		int count = Math.min((int)Math.ceil((double)(barWidth - 2) * current / maximum), barWidth - 2);
		if (count == shown)
			return;
		shown = count;
		
		StringBuilder str = new StringBuilder("\r");
		int i = 0;
		for (int j = 0; j < message.length(); j++) {
			char chr = message.charAt(j);
			int width = (chr > 0xFF ? 2 : 1);
			if (i + width > messageWidth)
				break;
			str.append(chr);
			i += width;
		}
		for (; i <= messageWidth; i++)
			str.append(' ');
		str.append('|');
		for (i = 0; i < count; i++)
			str.append('*');
		for (; i < barWidth - 2; i++)
			str.append(' ');
		str.append('|');
		System.err.print(str);
	}
}
