/*
	A sample sorting program for CTAL_0
 */

// Declare that we use externally supplied printf() function without checking its safety
__native__ int printf(const char *, ...);

// Simplified implementations of some standard functions
typedef int size_t;
static void *malloc(size_t size)
{
	return __malloc__(size);  // special malloc function provided by the compiler
}
static void *memcpy(void *dest, const void *src, size_t length)
{
	// Attempt to perform a wordwise copy, since a simple bytewise copy will discard
	// all base-pointers and make it impossible to use words contained in dest as pointers.
	char *dp = (char *)dest, *sp = (char *)src;
	int dp_align = (int)dest % sizeof(int), sp_align = (int)sp % sizeof(int);
	if (dp_align == sp_align) {
		while (length > 0 && dp_align-- > 0)
			*dp++ = *sp++, length--;
		while (length >= sizeof(int))
			*(int *)dp = *(int *)sp, dp += sizeof(int), sp += sizeof(int), length -= sizeof(int);
	}
	while (length > 0)
		*dp++ = *sp++, length--;
	return dest;
}
static int strlen(const char *str)
{
	int i = 0;
	while (*str++)
		i++;
	return i;
}
static char *strcpy(char *dest, const char *src)
{
	return memcpy(dest, src, strlen(src) + 1);
}
static char *strcat(char *str1, const char *str2)
{
	memcpy(str1 + strlen(str1), str2, strlen(str2) + 1);
	return str1;
}
static int puts(char *str)
{
	while (*str)
		printf("%c", *str++);
	printf("\n");
}

// Emulate sprintf(str, "%d", num) for non-negative num
static char *itoa(int num, char *str)
{
	int i = 0, j, k;
	while (num || !i)
		str[i++] = '0' + (num % 10), num /= 10;
	str[i--] = '\0';
	for (j = 0; j < i; i--, j++)
		k = str[j], str[j] = str[i], str[i] = k;
	return str;
}

// Insertion sort (similar to the standard qsort() function)
static void *isort(void *base, size_t nmemb, size_t size, int (*compar)(const void *, const void *))
{
	int i, j, k;
	void *swap = malloc(size);
	for (i = 1; i < nmemb; i++) {
		for (j = i; j > 0; j--) {
			void *prev = ((char *)base) + (j - 1) * size, *next = ((char *)base) + j * size;
			if (compar(prev, next) <= 0)
				break;
			memcpy(swap, prev, size);
			memcpy(prev, next, size);
			memcpy(next, swap, size);
		}
	}
	return base;
}

// A data structure and its comparator
struct point {
	int x, y;
	char *name;
};
static int compare_point(const void *o1, const void *o2)
{
	struct point *p1 = (struct point *)o1, *p2 = (struct point *)o2;
	if (p1->x != p2->x)
		return p1->x - p2->x;
	return p1->y - p2->y;
}

// Main
int main(void)
{
	struct point data[10][10];
	int i, j, k = 0;
	for (i = 0; i < 10; i++) {
		for (j = 0; j < 10; j++) {
			char buf[32];
			data[i][j].x = 10 - j;
			data[i][j].y = i;
			data[i][j].name = malloc(11);  // if you change this to 10, you will get a runtime error
			strcat(strcpy(data[i][j].name, "Point("), itoa(i, buf));
			strcat(strcat(strcat(data[i][j].name, ","), itoa(j, buf)), ")");
		}
	}
	isort(data, 10 * 10, sizeof(struct point), compare_point);
	for (i = 0; i < 10; i++) {
		for (j = 0; j < 10; j++) {
			printf("[%d] = %d, %d, ", k++, data[i][j].x, data[i][j].y);
			puts(data[i][j].name);
		}
	}
	return 0;
}
